######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Importer properties widget."""
from PySide6.QtCore import QPoint, Slot
from spinetoolbox.widgets.properties_widget import PropertiesWidgetBase
from ..item_info import ItemInfo
from .custom_menus import FilesContextMenu


class ImporterPropertiesWidget(PropertiesWidgetBase):
    """Widget for the Importer Item Properties."""

    def __init__(self, toolbox):
        """
        Args:
            toolbox (ToolboxUI): The toolbox instance where this widget should be embedded
        """
        super().__init__(toolbox)
        from ..ui.importer_properties import Ui_Form  # pylint: disable=import-outside-toplevel

        self.ui = Ui_Form()
        self.ui.setupUi(self)
        # Class attributes
        self.files_context_menu = None
        model = self._toolbox.filtered_spec_factory_models[ItemInfo.item_type()]
        self.ui.comboBox_specification.setModel(model)
        self.connect_signals()

    def connect_signals(self):
        """Connect signals to slots."""
        self.ui.treeView_files.customContextMenuRequested.connect(self.show_files_context_menu)

    @Slot(QPoint)
    def show_files_context_menu(self, pos):
        """Create and show a context-menu in Importer properties source files view.

        Args:
            pos (QPoint): Mouse position
        """
        ind = self.ui.treeView_files.indexAt(pos)  # Index of selected item in references tree view.
        importer = self._active_item
        global_pos = self.ui.treeView_files.viewport().mapToGlobal(pos)
        self.files_context_menu = FilesContextMenu(self, global_pos, ind)
        option = self.files_context_menu.get_action()
        if option == "Open import editor...":
            importer.open_import_editor(ind)
        elif option == "Open directory...":
            importer.open_directory()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains model for export mapping setup table."""
from enum import IntEnum, unique
from operator import itemgetter
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QColor, QFont
from spinedb_api.export_mapping.export_mapping import (
    AlternativeDescriptionMapping,
    AlternativeMapping,
    DefaultValueIndexNameMapping,
    DimensionMapping,
    ElementMapping,
    EntityClassMapping,
    EntityGroupEntityMapping,
    EntityGroupMapping,
    EntityMapping,
    ExpandedParameterDefaultValueMapping,
    ExpandedParameterValueMapping,
    FixedValueMapping,
    IndexNameMapping,
    ParameterDefaultValueIndexMapping,
    ParameterDefaultValueMapping,
    ParameterDefaultValueTypeMapping,
    ParameterDefinitionMapping,
    ParameterValueIndexMapping,
    ParameterValueListMapping,
    ParameterValueListValueMapping,
    ParameterValueMapping,
    ParameterValueTypeMapping,
    ScenarioAlternativeMapping,
    ScenarioBeforeAlternativeMapping,
    ScenarioDescriptionMapping,
    ScenarioMapping,
)
from spinedb_api.mapping import Position, is_pivoted, is_regular, value_index
from spinetoolbox.helpers import color_from_index, plain_to_rich
from ..commands import SetMappingNullable, SetMappingPositions, SetMappingProperty

POSITION_DISPLAY_TEXT = {Position.hidden: "hidden", Position.table_name: "table name", Position.header: "column header"}


@unique
class EditorColumn(IntEnum):
    ROW_LABEL = 0
    POSITION = 1
    PIVOTED = 2
    NULLABLE = 3
    HEADER = 4
    FILTER = 5


class MappingEditorTableModel(QAbstractTableModel):
    MAPPING_ITEM_ROLE = Qt.ItemDataRole.UserRole + 1

    def __init__(self, mapping_name, root_mapping, undo_stack, mapping_provider, parent=None):
        """
        Args:
            mapping_name (str): mapping's name
            root_mapping (ExportMapping, optional): root mapping
            undo_stack (QUndoStack): undo stack
            mapping_provider (SpecificationEditorWindow): window that can provide data for different mappings
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._root_mapping = None
        self._mappings = []
        self._mapping_colors = {}
        self._reset_mapping(root_mapping)
        self._non_leaf_mapping_is_pivoted = self._is_non_leaf_pivoted()
        self._undo_stack = undo_stack
        self._mapping_name = mapping_name
        self._mapping_provider = mapping_provider

    def _reset_mapping(self, root_mapping):
        """Sets the root mapping attribute, then updates mappings and colors."""
        self._root_mapping = root_mapping
        if self._root_mapping is None:
            self._mappings = []
            self._mapping_colors = {}
            return
        self._mappings = self._root_mapping.flatten()
        if self._mappings and isinstance(self._mappings[0], FixedValueMapping):
            # Pop the first element if it's a FixedValueMapping. We don't want to have the fixed table name here.
            self._mappings.pop(0)
        self._reset_colors(emit_data_changed=False)

    def _reset_colors(self, emit_data_changed=True):
        """Recalculates position column background colors.

        Args:
            emit_data_changed (boo): if True emits relevant dataChanged signal
        """
        positions = [m.position for m in self._mappings if isinstance(m.position, int)]
        self._mapping_colors = {p: color_from_index(i, len(positions)).lighter() for i, p in enumerate(positions)}
        if emit_data_changed:
            top_left = self.index(0, EditorColumn.POSITION)
            bottom_right = self.index(len(self._mappings), EditorColumn.POSITION)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.BackgroundRole])

    def mapping_colors(self):
        return self._mapping_colors

    def columnCount(self, paren=QModelIndex()):
        return len(EditorColumn)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        column = index.column()
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            row = index.row()
            if column == EditorColumn.ROW_LABEL:
                mapping = self._mappings[row]
                return _names[type(mapping)]
            if column == EditorColumn.POSITION:
                if row == value_index(self._mappings) and self._non_leaf_mapping_is_pivoted:
                    return "in pivot"
                position = self._mappings[row].position
                if is_regular(position):
                    return str(position + 1)
                if is_pivoted(position):
                    return str(-position)
                return POSITION_DISPLAY_TEXT.get(position, "unrecognized")
            if column == EditorColumn.HEADER:
                return self._mappings[row].header
            if column == EditorColumn.FILTER:
                return self._mappings[row].filter_re
        elif role == Qt.ItemDataRole.CheckStateRole:
            if column == EditorColumn.PIVOTED:
                if is_pivoted(self._mappings[index.row()].position):
                    return Qt.CheckState.Checked
                return Qt.CheckState.Unchecked
            if column == EditorColumn.NULLABLE:
                return Qt.CheckState.Checked if self._mappings[index.row()].is_ignorable() else Qt.CheckState.Unchecked
        elif role == Qt.ItemDataRole.FontRole and column == EditorColumn.ROW_LABEL:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.ItemDataRole.BackgroundRole and column == EditorColumn.ROW_LABEL:
            m = self._mappings[index.row()]
            return self._mapping_colors.get(m.position, QColor(Qt.GlobalColor.gray).lighter())
        elif role == Qt.ItemDataRole.ToolTipRole:
            if column == EditorColumn.FILTER:
                return plain_to_rich("Regular expression to filter database items.")
            if column == EditorColumn.NULLABLE:
                return plain_to_rich("When checked, ignore this row if it yields nothing to export.")
        if role == self.MAPPING_ITEM_ROLE:
            return self._mappings[index.row()]
        return None

    def flags(self, index=QModelIndex()):
        row = index.row()
        column = index.column()
        value_row = value_index(self._mappings)
        if row >= value_row and column == EditorColumn.PIVOTED:
            return super().flags(index) & ~Qt.ItemIsEnabled
        if column == EditorColumn.ROW_LABEL:
            return super().flags(index) & ~Qt.ItemIsSelectable
        if column in (EditorColumn.POSITION, EditorColumn.FILTER):
            return super().flags(index) | Qt.ItemIsEditable
        if column == EditorColumn.PIVOTED:
            return super().flags(index) | Qt.ItemIsUserCheckable
        if column == EditorColumn.HEADER:
            if self._mappings[row].position == Position.header:
                return super().flags(index) & ~Qt.ItemIsEnabled
            return super().flags(index) | Qt.ItemIsEditable
        if column == EditorColumn.NULLABLE:
            return super().flags(index) | Qt.ItemIsUserCheckable
        return super().flags(index)

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return ("Mapping type", "Map to", "Pivoted", "Nullable", "Header", "Filter")[section]
        return None

    def root_mapping(self):
        """
        Returns:
            Mapping: root mapping
        """
        return self._root_mapping

    def rowCount(self, parent=QModelIndex()):
        return len(self._mappings)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        column = index.column()
        if role == Qt.ItemDataRole.EditRole:
            row = index.row()
            mapping = self._mappings[row]
            if column == EditorColumn.POSITION:
                return self._push_set_positions_command(value, row, mapping)
            if column == EditorColumn.HEADER:
                if value == mapping.header:
                    return False
                previous_header = mapping.header
                command = SetMappingProperty(
                    "change mapping item's header", self.set_header, self._mapping_name, row, value, previous_header
                )
                self._undo_stack.push(command)
                return True
            if column == EditorColumn.FILTER:
                if value == mapping.filter_re:
                    return False
                previous_filter_re = mapping.filter_re
                command = SetMappingProperty(
                    "change mapping item's filter",
                    self.set_filter_re,
                    self._mapping_name,
                    row,
                    value,
                    previous_filter_re,
                )
                self._undo_stack.push(command)
                return True
        elif role == Qt.ItemDataRole.CheckStateRole:
            if column == EditorColumn.PIVOTED:
                row = index.row()
                new_positions = _propose_toggled_pivot(self._mappings, row)
                previous_positions = [m.position for m in self._mappings]
                command = SetMappingPositions(self, self._mapping_name, new_positions, previous_positions)
                self._undo_stack.push(command)
                return True
            if column == EditorColumn.NULLABLE:
                row = index.row()
                nullable = self._mappings[row].is_ignorable()
                self._undo_stack.push(SetMappingNullable(self, self._mapping_name, row, not nullable))
                return True
        return False

    def set_mapping(self, mapping_name, root_mapping):
        """
        Sets mapping for model.

        Args:
            mapping_name (str): mapping's name
            root_mapping (Mapping, optional): root mapping
        """
        self.beginResetModel()
        self._mapping_name = mapping_name
        self._reset_mapping(root_mapping)
        self._non_leaf_mapping_is_pivoted = self._is_non_leaf_pivoted()
        self.endResetModel()

    def _push_set_positions_command(self, value, row, mapping):
        """Pushes SetMappingPosition command to undo stack

        Args:
            value (str): table cell's value
            row (int): row index
            mapping (Mapping): mapping to modify
        """
        try:
            value = max(int(value), 1)
            if is_pivoted(mapping.position):
                value = -value
            else:
                value = value - 1
        except ValueError:
            value = value.strip().lower()
            if value.startswith("t"):
                value = Position.table_name
            elif value.startswith("h"):
                value = Position.hidden
            elif value.startswith("c"):
                value = Position.header
            else:
                return False
        if value == mapping.position and (row != value_index(self._mappings) or not self._non_leaf_mapping_is_pivoted):
            return False
        positions = _propose_positions(self._mappings, row, value)
        previous_positions = [m.position for m in self._mappings]
        command = SetMappingPositions(self, self._mapping_name, positions, previous_positions)
        self._undo_stack.push(command)
        return True

    def positions(self):
        """Returns mapping positions.

        Returns:
            list: positions
        """
        return [m.position for m in self._mappings]

    def set_positions(self, positions, mapping_name):
        """
        Sets mapping position for given row.

        Args:
            positions (list of Position): mapping position
            mapping_name (str): mapping's name
        """
        if mapping_name != self._mapping_name:
            self._mapping_provider.show_on_table(mapping_name)
        top = -1
        pivot_top = top
        bottom = len(self._mappings)
        pivot_bottom = bottom
        for row, (mapping, position) in enumerate(zip(self._mappings, positions)):
            if position != mapping.position:
                top = row
                bottom = min(bottom, row)
                if is_pivoted(position) or is_pivoted(mapping.position):
                    pivot_top = row
                    pivot_bottom = min(pivot_bottom, row)
                mapping.position = position
        top_left = self.index(top, EditorColumn.POSITION)
        bottom_right = self.index(bottom, EditorColumn.POSITION)
        self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])
        if pivot_bottom <= pivot_top:
            top_left = self.index(pivot_top, EditorColumn.PIVOTED)
            bottom_right = self.index(pivot_bottom, EditorColumn.PIVOTED)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.CheckStateRole])
        non_leaf_pivoted = self._is_non_leaf_pivoted()
        if non_leaf_pivoted != self._non_leaf_mapping_is_pivoted:
            self._non_leaf_mapping_is_pivoted = non_leaf_pivoted
            index = self.index(value_index(self._mappings), EditorColumn.POSITION)
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])
        self._reset_colors()

    def set_header(self, header, row, mapping_name):
        """
        Sets mapping header for given row.

        Args:
            header (str): mapping header
            row (int): row index
            mapping_name (str): mapping's name
        """
        if mapping_name != self._mapping_name:
            self._mapping_provider.show_on_table(mapping_name)
        self._mappings[row].header = header
        index = self.index(row, EditorColumn.HEADER)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])

    def set_filter_re(self, filter_re, row, mapping_name):
        """
        Sets mapping filter_re for given row.

        Args:
            filter_re (str): mapping filter_re
            row (int): row index
            mapping_name (str): mapping's name
        """
        if mapping_name != self._mapping_name:
            self._mapping_provider.show_on_table(mapping_name)
        self._mappings[row].filter_re = filter_re
        index = self.index(row, EditorColumn.FILTER)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole])

    def set_nullable(self, nullable, row, mapping_name):
        """
        Sets mapping filter_re for given row.

        Args:
            nullable (bool): mapping filter_re
            row (int): row index
            mapping_name (str): mapping's name
        """
        if mapping_name != self._mapping_name:
            self._mapping_provider.show_on_table(mapping_name)
        self._mappings[row].set_ignorable(nullable)
        index = self.index(row, EditorColumn.NULLABLE)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.CheckStateRole])

    def _is_non_leaf_pivoted(self):
        """Checks if a non-leaf mapping is pivoted.

        Returns:
            bool: True if one or more non-leaf mappings are pivoted, False otherwise
        """
        return len(self._mappings) > 1 and any(
            is_pivoted(m.position) for m in self._mappings[: value_index(self._mappings)]
        )

    def can_compact(self):
        """Checks if the mapping can be compacted.

        Returns:
            bool: True if the mapping can be compacted, False otherwise
        """
        regular_positions = sorted([m.position for m in self._mappings if is_regular(m.position)])
        for i, position in enumerate(regular_positions):
            if i != position:
                return True
        pivoted_positions = sorted([m.position for m in self._mappings if is_pivoted(m.position)])
        for i, position in enumerate(pivoted_positions):
            if i != position:
                return True
        return False

    def compact(self):
        """Compacts the mapping."""
        if self._is_non_leaf_pivoted():
            value_row = value_index(self._mappings)
            pivoted_mappings = [
                (m.position, m) for row, m in enumerate(self._mappings) if is_pivoted(m.position) and row != value_row
            ]
            pivoted_mappings.sort(reverse=True, key=itemgetter(0))
            for row, item in enumerate(pivoted_mappings):
                item[1].position = -(row + 1)
            regular_mappings = [
                (m.position, m) for row, m in enumerate(self._mappings) if is_regular(m.position) and row != value_row
            ]
            last_column = max(regular_mappings, key=itemgetter(0), default=-1)[0]
            regular_mappings.append((last_column + 1, self._mappings[value_row]))
        else:
            regular_mappings = [(m.position, m) for m in self._mappings if is_regular(m.position)]
        regular_mappings.sort(key=itemgetter(0))
        for column, item in enumerate(regular_mappings):
            item[1].position = column
        top_left = self.index(0, EditorColumn.POSITION)
        bottom_right = self.index(self.rowCount() - 1, EditorColumn.POSITION)
        self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])
        self._reset_colors()


_names = {
    AlternativeDescriptionMapping: "Alternatives description",
    AlternativeMapping: "Alternatives",
    DefaultValueIndexNameMapping: "Default value index names",
    ExpandedParameterDefaultValueMapping: "Default values",
    ExpandedParameterValueMapping: "Parameter values",
    IndexNameMapping: "Parameter index names",
    EntityClassMapping: "Entity classes",
    EntityGroupMapping: "Entity groups",
    EntityGroupEntityMapping: "Entities",
    EntityMapping: "Entities",
    ElementMapping: "Elements",
    ParameterDefaultValueMapping: "Default values",
    ParameterDefaultValueIndexMapping: "Default value indexes",
    ParameterDefaultValueTypeMapping: "Default value types",
    ParameterDefinitionMapping: "Parameter definitions",
    ParameterValueIndexMapping: "Parameter indexes",
    ParameterValueListMapping: "Value lists",
    ParameterValueListValueMapping: "Value list values",
    ParameterValueMapping: "Parameter values",
    ParameterValueTypeMapping: "Value types",
    DimensionMapping: "Dimensions",
    ScenarioAlternativeMapping: "Alternatives",
    ScenarioBeforeAlternativeMapping: "Before alternatives",
    ScenarioDescriptionMapping: "Scenarios description",
    ScenarioMapping: "Scenarios",
}


def _propose_toggled_pivot(mappings, toggled_index):
    """Proposes new positions after toggling a mapping's pivoted status.

    Args:
        mappings (list of Mapping): flattened mappings
        toggled_index (int): toggled index in mappings

    Returns:
        list of Position: positions after toggling
    """
    positions = [m.position for m in mappings]
    previous_position = positions[toggled_index]
    if previous_position in (Position.hidden, Position.table_name, Position.header):
        previous_position = 0
    new_position = -previous_position - 1
    if new_position < 0:
        positions[toggled_index] = new_position
        _remove_column(positions, previous_position)
    else:
        _insert_into_position(positions, toggled_index, new_position)
    return positions


def _propose_positions(mappings, target_index, new_position):
    """Proposes new positions.

    Args:
        mappings (list of Mapping): flattened mappings
        target_index (int): index of mapping to modify
        new_position (Position or int): mapping's new position

    Returns:
        list of Position: positions after modification
    """
    positions = [m.position for m in mappings]
    if target_index == value_index(mappings):
        _turn_off_pivots(positions)
    if isinstance(new_position, int):
        _insert_into_position(positions, target_index, new_position)
    else:
        if new_position == Position.table_name:
            try:
                other_table_name_index = positions.index(Position.table_name)
            except ValueError:
                pass
            else:
                positions[other_table_name_index] = positions[target_index]
        positions[target_index] = new_position
    return positions


def _turn_off_pivots(positions):
    """Toggles pivoted positions.

    Args:
        positions (list of Position): proposed positions
    """
    for row, position in enumerate(positions):
        if is_pivoted(position):
            new_position = -position - 1
            _insert_into_position(positions, row, new_position)


def _insert_into_position(positions, target_index, new_position):
    """Inserts a position.

    Args:
        positions (list of Position): proposed positions
        target_index (int): index of position to modify
        new_position (Position or int): new position
    """
    for i in range(len(positions)):  # pylint: disable=consider-using-enumerate
        if i == target_index:
            continue
        if positions[i] == new_position:
            direction = 1 if new_position >= 0 else -1
            _insert_into_position(positions, i, new_position + direction)
            break
    positions[target_index] = new_position


def _remove_column(positions, column):
    """Removes a regular column position.

    Args:
        positions (list of Position): proposed positions

    """
    new_position = column + 1
    for i in range(len(positions)):  # pylint: disable=consider-using-enumerate
        if positions[i] == new_position:
            _insert_into_position(positions, i, new_position)
            break



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""A small widget to set up an output in Exporter properties tab."""
from PySide6.QtCore import Signal, Slot
from PySide6.QtWidgets import QWidget
from spine_items.utils import convert_to_sqlalchemy_url
from spine_items.widgets import UrlSelectorDialog


class ExportListItem(QWidget):
    """A widget with few controls to select the export output label."""

    _URL_SET_TEXT = "Output database URL set."
    _URL_UNSET_TEXT = "No output database URL set."
    _URL_IGNORED_TEXT = "Ignoring output database URL."

    out_label_changed = Signal(str, str)
    """Emitted when the output label field changes."""
    out_label_editing_finished = Signal(str)
    """Emitted when output label editing is finished."""
    out_url_changed = Signal(str, dict)
    """Emitted when the output URL changes."""

    def __init__(self, in_label, out_label, out_url, app_settings, logger, parent=None):
        """
        Args:
            in_label (str): input resource label
            out_label (str): output resource label
            out_url (dict): output ULR
            app_settings (QSettings): Toolbox settings
            logger (LoggerInterface): logger
            parent (QWidget): a parent widget
        """
        from ..ui.export_list_item import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._ui = Ui_Form()
        self._in_label = in_label
        self._out_label = out_label
        self._out_url = out_url
        self._app_settings = app_settings
        self._logger = logger
        self._ui.setupUi(self)
        self.set_in_label(in_label)
        self._ui.out_url_status_label.setText(self._URL_UNSET_TEXT)
        self._ui.out_label_edit.setText(out_label)
        self._ui.out_label_edit.textEdited.connect(self._emit_out_label_changed)
        self._ui.out_label_edit.editingFinished.connect(self._emit_out_label_editing_finished)
        self._ui.out_url_setup_button.clicked.connect(self._show_url_dialog)
        self._ui.out_url_clear_button.clicked.connect(self._clear_out_url)

    @property
    def out_label_edit(self):
        """Output label QLineEdit"""
        return self._ui.out_label_edit

    @property
    def in_label_field(self):
        """Input label QLineEdit"""
        return self._ui.input_label_field

    def set_in_label(self, in_label):
        """Sets new input label.

        Args:
            in_label (str): input label
        """
        self._ui.input_label_field.setText(f"<b>{in_label}</b>")

    def set_out_url_enabled(self, enabled):
        """Enables or disables the output URL widgets.

        Args:
            enabled (bool): True to enable, False to disable
        """
        if not enabled:
            self._ui.out_url_status_label.setText(self._URL_IGNORED_TEXT)
        else:
            self._update_url_dependent_widgets()
        self._ui.out_url_widget.setEnabled(enabled)

    def set_out_url(self, url_dict):
        """Sets new output URL.

        Args:
            url_dict (dict, optional): URL dict
        """
        self._out_url = url_dict
        self._update_url_dependent_widgets()

    def _update_url_dependent_widgets(self):
        """Sets suitable text for the status label and enables the clear button as needed."""
        sa_url = convert_to_sqlalchemy_url(self._out_url)
        if sa_url is not None:
            self._ui.out_url_clear_button.setEnabled(True)
            self._ui.out_url_status_label.setText(self._URL_SET_TEXT)
        else:
            self._ui.out_url_clear_button.setEnabled(False)
            self._ui.out_url_status_label.setText(self._URL_UNSET_TEXT)

    @Slot(str)
    def _emit_out_label_changed(self, out_label):
        """Emits out_label_changed signal.

        Args:
            out_label (str): new out label
        """
        if self._out_label == out_label:
            return
        self._out_label = out_label
        self.out_label_changed.emit(out_label, self._in_label)

    @Slot()
    def _emit_out_label_editing_finished(self):
        self.out_label_editing_finished.emit(self._in_label)

    @Slot(bool)
    def _show_url_dialog(self, _=False):
        """Opens the URL selector dialog."""
        dialog = UrlSelectorDialog(self._app_settings, True, self._logger, self)
        if self._out_url is not None:
            dialog.set_url_dict(self._out_url)
        dialog.exec_()
        url = dialog.url_dict()
        if not url:
            return
        self.set_out_url(url)
        self.out_url_changed.emit(self._in_label, self._out_url)

    @Slot(bool)
    def _clear_out_url(self, _=False):
        """Clears the output URL."""
        self.set_out_url(None)
        self.out_url_changed.emit(self._in_label, self._out_url)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ''specification'' module"""
import unittest
from spine_items.exporter.specification import MappingSpecification, MappingType, OutputFormat, Specification
from spinedb_api.export_mapping import entity_export
from spinedb_api.mapping import Position


class TestSpecification(unittest.TestCase):
    def test_is_exporting_multiple_files_is_false_when_exporting_single_file(self):
        mapping_root = entity_export(0, 1)
        mapping_specification = MappingSpecification(MappingType.entities, True, False, "", False, mapping_root)
        mapping_specifications = {"Only mapping": mapping_specification}
        specification = Specification(mapping_specifications=mapping_specifications, output_format=OutputFormat.CSV)
        self.assertFalse(specification.is_exporting_multiple_files())

    def test_is_exporting_multiple_files_is_false_when_exporting_csv_with_named_tables(self):
        mapping_root = entity_export(Position.table_name, 0)
        mapping_specification = MappingSpecification(MappingType.entities, True, False, "", False, mapping_root)
        mapping_specifications = {"Only mapping": mapping_specification}
        specification = Specification(mapping_specifications=mapping_specifications, output_format=OutputFormat.CSV)
        self.assertTrue(specification.is_exporting_multiple_files())


class TestOutputFormat(unittest.TestCase):
    def test_compatible_file_extensions(self):
        self.assertTrue(OutputFormat.CSV.is_compatible_file_extension("csv"))
        self.assertTrue(OutputFormat.CSV.is_compatible_file_extension("dat"))
        self.assertTrue(OutputFormat.CSV.is_compatible_file_extension("txt"))
        self.assertFalse(OutputFormat.CSV.is_compatible_file_extension("XXX"))
        self.assertTrue(OutputFormat.EXCEL.is_compatible_file_extension("xlsx"))
        self.assertFalse(OutputFormat.EXCEL.is_compatible_file_extension("XXX"))
        self.assertTrue(OutputFormat.GDX.is_compatible_file_extension("gdx"))
        self.assertFalse(OutputFormat.GDX.is_compatible_file_extension("XXX"))
        self.assertTrue(OutputFormat.SQL.is_compatible_file_extension("sqlite"))
        self.assertFalse(OutputFormat.SQL.is_compatible_file_extension("XXX"))

    def test_every_format_has_file_extensions(self):
        for output_format in OutputFormat:
            try:
                output_format.file_extension()
            except KeyError:
                self.fail(f"{output_format} is missing file extension")

    def test_output_format_from_extension(self):
        self.assertEqual(OutputFormat.output_format_from_extension("csv"), OutputFormat.CSV)
        self.assertEqual(OutputFormat.output_format_from_extension("dat"), OutputFormat.CSV)
        self.assertEqual(OutputFormat.output_format_from_extension("txt"), OutputFormat.CSV)
        self.assertEqual(OutputFormat.output_format_from_extension("gdx"), OutputFormat.GDX)
        self.assertEqual(OutputFormat.output_format_from_extension("sqlite"), OutputFormat.SQL)
        self.assertEqual(OutputFormat.output_format_from_extension("xlsx"), OutputFormat.EXCEL)
        self.assertIsNone(OutputFormat.output_format_from_extension("XXX"))

    def test_default_format(self):
        self.assertEqual(OutputFormat.default(), OutputFormat.CSV)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes for custom context menus and pop-up menus."""
from spinetoolbox.widgets.custom_menus import CustomContextMenu


class ViewRefsContextMenu(CustomContextMenu):
    """Context menu class for the references tree view of the View project item properties."""

    def __init__(self, parent, position, view):
        """

        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            view (View): View item that requests the menu
        """
        super().__init__(parent, position)
        selected_references = view.selected_references()
        self.add_action("Pin values...", enabled=bool(selected_references))
        self.add_action("Open editor...", enabled=bool(selected_references))


class ViewSelectionsContextMenu(CustomContextMenu):
    """Context menu class for the pinned values tree view of the View project item properties."""

    def __init__(self, parent, position, view):
        """

        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            view (View): View item that requests the menu
        """
        super().__init__(parent, position)
        selected_pinned_values = view.selected_pinned_values()
        self.add_action("Plot", enabled=bool(selected_pinned_values))
        self.add_action("Copy plot data", enabled=len(selected_pinned_values) == 1)
        self.addSeparator()
        self.add_action("Unpin", enabled=bool(selected_pinned_values))
        self.add_action("Rename...", enabled=len(selected_pinned_values) == 1)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Module for Merger class."""
import os
from PySide6.QtCore import Qt, Slot
from spinetoolbox.helpers import create_dir
from ..commands import UpdateCancelOnErrorCommand
from ..db_writer_item_base import DBWriterItemBase
from .executable_item import ExecutableItem
from .item_info import ItemInfo


class Merger(DBWriterItemBase):
    def __init__(self, name, description, x, y, toolbox, project, cancel_on_error=False):
        """Data Store class.

        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): the project this item belongs to
            cancel_on_error (bool): if True, changes will be reverted on errors
        """
        super().__init__(name, description, x, y, project)
        self._toolbox = toolbox
        self.logs_dir = os.path.join(self.data_dir, "logs")
        try:
            create_dir(self.logs_dir)
        except OSError:
            self._logger.msg_error.emit(f"[OSError] Creating directory {self.logs_dir} failed. Check permissions.")
        self.cancel_on_error = cancel_on_error

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        s[self._properties_ui.cancel_on_error_checkBox.stateChanged] = self._handle_cancel_on_error_changed
        return s

    def restore_selections(self):
        """Load url into selections."""
        self._properties_ui.cancel_on_error_checkBox.setCheckState(
            Qt.CheckState.Checked if self.cancel_on_error else Qt.CheckState.Unchecked
        )

    @Slot(int)
    def _handle_cancel_on_error_changed(self, _state):
        cancel_on_error = self._properties_ui.cancel_on_error_checkBox.isChecked()
        if self.cancel_on_error == cancel_on_error:
            return
        self._toolbox.undo_stack.push(UpdateCancelOnErrorCommand(self.name, cancel_on_error, self._project))

    def set_cancel_on_error(self, cancel_on_error):
        self.cancel_on_error = cancel_on_error
        if not self._active:
            return
        check_state = Qt.CheckState.Checked if self.cancel_on_error else Qt.CheckState.Unchecked
        self._properties_ui.cancel_on_error_checkBox.blockSignals(True)
        self._properties_ui.cancel_on_error_checkBox.setCheckState(check_state)
        self._properties_ui.cancel_on_error_checkBox.blockSignals(False)

    def predecessor_data_stores(self):
        for name in self._project.predecessor_names(self.name):
            item = self._project.get_item(name)
            if item.item_type() == "Data Store":
                yield item

    def predecessor_data_transformers(self):
        for name in self._project.predecessor_names(self.name):
            item = self._project.get_item(name)
            if item.item_type() == "Data Transformer":
                yield item

    def upstream_resources_updated(self, resources):
        self._check_notifications()

    def downstream_resources_updated(self, resources):
        self._check_notifications()

    def _check_notifications(self):
        self.clear_notifications()
        self._check_write_index()
        if not (list(self.predecessor_data_stores()) or list(self.predecessor_data_transformers())):
            self.add_notification(
                "This Merger does not have any input Data Stores or Data Transformers. "
                "Connect Data Stores/Transformers to this to merge their data into output Data Stores/Transformers."
            )
        if not list(self.successor_data_stores()):
            self.add_notification(
                "This Merger does not have any output Data Stores. "
                "Connect this to Data Stores to merge input Data Stores/Transformers data into them."
            )

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        d["cancel_on_error"] = self.cancel_on_error
        return d

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        """See base class."""
        description, x, y = DBWriterItemBase.parse_item_dict(item_dict)
        cancel_on_error = item_dict.get("cancel_on_error", False)
        return Merger(name, description, x, y, toolbox, project, cancel_on_error)

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Data Store":
            dst_ds_names = ", ".join(x.name for x in self.successor_data_stores())
            if dst_ds_names:
                self._logger.msg.emit(
                    "Link established. "
                    f"Data from <b>{source_item.name}</b> will be merged into <b>{dst_ds_names}</b> upon execution."
                )
            else:
                self._logger.msg.emit("Link established. " f"<b>{self.name} is missing output database, though.</b>")
        else:
            super().notify_destination(source_item)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`MultiCheckableListView`."""
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QListView


class MultiCheckableListView(QListView):
    """A list view which allows all selected items to be checked/unchecked with space bar."""

    def keyPressEvent(self, event):
        """Handles key press events."""
        if event.key() != Qt.Key_Space or event.modifiers() != Qt.NoModifier:
            super().keyPressEvent(event)
            return
        selection_model = self.selectionModel()
        if not selection_model.hasSelection():
            super().keyPressEvent(event)
            return
        selected = selection_model.selectedIndexes()
        model = self.model()
        if len(selected) == 1:
            state = selected[0].data(Qt.ItemDataRole.CheckStateRole)
            check_state = Qt.CheckState.Checked if state == Qt.CheckState.Unchecked else Qt.CheckState.Unchecked
            model.setData(selected[0], check_state.value, Qt.ItemDataRole.CheckStateRole)
        else:
            model.toggle_checked_tables(selected)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for export mapping setup table."""
import unittest
from unittest.mock import MagicMock
from PySide6.QtGui import QUndoStack
from PySide6.QtWidgets import QApplication
from spine_items.exporter.mvcmodels.mapping_editor_table_model import MappingEditorTableModel
from spinedb_api.export_mapping import entity_export
from spinedb_api.mapping import Position


class TestMappingTableModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._undo_stack = QUndoStack()

    def test_columnCount(self):
        model = MappingEditorTableModel(
            "mapping", entity_export(Position.hidden, Position.hidden), self._undo_stack, MagicMock()
        )
        self.assertEqual(model.rowCount(), 2)

    def test_rowCount(self):
        mapping_root = entity_export(Position.hidden, Position.hidden)
        model = MappingEditorTableModel("mapping", mapping_root, self._undo_stack, MagicMock())
        self.assertEqual(model.rowCount(), mapping_root.count_mappings())

    def test_data(self):
        model = MappingEditorTableModel("mapping", entity_export(1, 2), self._undo_stack, MagicMock())
        self.assertEqual(model.rowCount(), 2)
        self.assertEqual(model.index(0, 0).data(), "Entity classes")
        self.assertEqual(model.index(0, 1).data(), "2")
        self.assertEqual(model.index(1, 0).data(), "Entities")
        self.assertEqual(model.index(1, 1).data(), "3")

    def test_setData_column_number(self):
        model = MappingEditorTableModel(
            "mapping", entity_export(Position.hidden, Position.hidden), self._undo_stack, MagicMock()
        )
        self.assertTrue(model.setData(model.index(0, 1), "23"))
        self.assertEqual(model.index(0, 1).data(), "23")

    def test_setData_prevents_duplicate_table_name_positions(self):
        model = MappingEditorTableModel("mapping", entity_export(Position.table_name, 0), self._undo_stack, MagicMock())
        self.assertEqual(model.rowCount(), 2)
        self.assertEqual(model.index(0, 0).data(), "Entity classes")
        self.assertEqual(model.index(0, 1).data(), "table name")
        self.assertEqual(model.index(1, 0).data(), "Entities")
        self.assertEqual(model.index(1, 1).data(), "1")
        model.setData(model.index(1, 1), "table name")
        self.assertEqual(model.index(0, 0).data(), "Entity classes")
        self.assertEqual(model.index(0, 1).data(), "1")
        self.assertEqual(model.index(1, 0).data(), "Entities")
        self.assertEqual(model.index(1, 1).data(), "table name")


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Exporter project item."""
import gc
import os.path
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock, NonCallableMagicMock
from PySide6.QtWidgets import QApplication
from spine_engine.project_item.project_item_resource import database_resource
from spine_engine.utils.helpers import shorten
from spine_items.data_store.data_store import DataStore
from spine_items.exporter.exporter import Exporter
from spine_items.exporter.exporter_factory import ExporterFactory
from spine_items.exporter.item_info import ItemInfo
from spine_items.exporter.specification import OutputFormat, Specification
from spine_items.utils import database_label
from spinedb_api import create_new_spine_database
from spinetoolbox.project_item.logging_connection import LoggingConnection
from tests.mock_helpers import (
    clean_up_toolbox,
    create_mock_project,
    create_mock_toolbox,
    create_toolboxui_with_project,
    mock_finish_project_item_construction,
)


class TestExporter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        """Set up."""
        self._toolbox = create_mock_toolbox()
        mock_spec_model = self._toolbox.specification_model = MagicMock()
        mock_spec_model.find_specification.side_effect = lambda name: Specification(name, "spec description")
        factory = ExporterFactory()
        item_dict = {
            "type": "Exporter",
            "description": "item description",
            "specification": "exporter_spec",
            "cancel_on_error": True,
            "x": 0,
            "y": 0,
        }
        self._temp_dir = TemporaryDirectory()
        self._project = create_mock_project(self._temp_dir.name)
        self._toolbox.project.return_value = self._project
        self._exporter = factory.make_item("exporter", item_dict, self._toolbox, self._project)
        self._properties_widget = mock_finish_project_item_construction(factory, self._exporter, self._toolbox)

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(Exporter.item_type(), ItemInfo.item_type())

    def test_serialization(self):
        item_dict = self._exporter.item_dict()
        deserialized = Exporter.from_dict("new exporter", item_dict, self._toolbox, self._project)
        self.assertEqual(deserialized.name, "new exporter")
        self.assertEqual(deserialized.description, "item description")

    def test_notify_destination(self):
        self._exporter._toolbox.msg = MagicMock()
        self._exporter._toolbox.msg_warning = MagicMock()
        source_item = NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = MagicMock(return_value="Data Connection")
        self._exporter.notify_destination(source_item)
        self._exporter.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Data Connection</b> and a <b>Exporter</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Data Store")
        self._exporter.notify_destination(source_item)
        self._exporter._toolbox.msg.emit.assert_called_with(
            "Link established. You can now export the database in <b>source name</b> in <b>exporter</b>."
        )
        source_item.item_type = MagicMock(return_value="Data Transformer")
        self._exporter.notify_destination(source_item)
        self._exporter._toolbox.msg.emit.assert_called_with(
            "Link established. You can now export the database transformed by <b>source name</b> in <b>exporter</b>."
        )
        source_item.item_type = MagicMock(return_value="Exporter")
        self._exporter.notify_destination(source_item)
        self._exporter.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Exporter</b> and a <b>Exporter</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Importer")
        self._exporter.notify_destination(source_item)
        self._exporter.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Importer</b> and a <b>Exporter</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Tool")
        self._exporter.notify_destination(source_item)
        self._exporter.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>Tool</b> and a <b>Exporter</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="View")
        self._exporter.notify_destination(source_item)
        self._exporter.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>View</b> and a <b>Exporter</b> has not been implemented yet."
        )

    def test_rename(self):
        """Tests renaming an Importer."""
        self._exporter.activate()
        expected_name = "ABC"
        expected_short_name = shorten(expected_name)
        self.assertTrue(self._exporter.rename(expected_name, ""))
        self.assertEqual(expected_name, self._exporter.name)
        self.assertEqual(expected_name, self._exporter.get_icon().name())
        expected_data_dir = os.path.join(self._project.items_dir, expected_short_name)
        self.assertEqual(expected_data_dir, self._exporter.data_dir)

    def test_upstream_resources_updated(self):
        self._exporter.activate()
        expected_out_labels = ["db_url@Data Store 1", "db_url@Data Store 2"]
        resources = [database_resource("provider", url) for url in expected_out_labels]
        self._exporter.upstream_resources_updated(resources)
        model = self._exporter.full_url_model()
        self.assertEqual(model.rowCount(), len(expected_out_labels))
        self.assertIn(model.index(0, 0).data(), expected_out_labels)
        self.assertIn(model.index(1, 0).data(), expected_out_labels)
        self.assertEqual(self._exporter._properties_ui.outputs_list_layout.count(), len(expected_out_labels))


class TestExporterWithToolbox(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)

    def tearDown(self):
        clean_up_toolbox(self._toolbox)
        gc.collect()
        self._temp_dir.cleanup()

    def test_notifications_when_not_configured(self):
        project = self._toolbox.project()
        exporter = Exporter("Test exporter", "Exporter for unit testing", 0.0, 0.0, self._toolbox, project)
        project.add_item(exporter)
        self.assertEqual(exporter.get_icon().exclamation_icon._notifications, ["Export specification missing."])

    def test_notifications_when_specification_is_set(self):
        project = self._toolbox.project()
        specification = Specification("Test spec", "Exporter specification for testing")
        project.add_specification(specification)
        exporter = Exporter("Test exporter", "Exporter for unit testing", 0.0, 0.0, self._toolbox, project, "Test spec")
        project.add_item(exporter)
        self.assertEqual(exporter.get_icon().exclamation_icon._notifications, [])

    def test_notifications_when_output_file_name_extension_mismatches_with_specification_output_format(self):
        project = self._toolbox.project()
        url = "sqlite:///" + os.path.join(self._temp_dir.name, "db.sqlite")
        create_new_spine_database(url)
        url_dict = {"dialect": "sqlite", "database": os.path.join(self._temp_dir.name, "db.sqlite")}
        data_store = DataStore("Dummy data store", "", 0.0, 0.0, self._toolbox, project, url_dict)
        project.add_item(data_store)
        while not data_store.is_url_validated():
            QApplication.processEvents()
        specification = Specification("Test spec", "Exporter specification for testing", output_format=OutputFormat.CSV)
        project.add_specification(specification)
        exporter = Exporter("Test exporter", "Exporter for unit testing", 0.0, 0.0, self._toolbox, project, "Test spec")
        project.add_item(exporter)
        connection = LoggingConnection("Dummy data store", "right", "Test exporter", "left", toolbox=self._toolbox)
        project.add_connection(connection)
        self.assertEqual(exporter.get_icon().exclamation_icon._notifications, [])
        exporter.activate()
        exporter.set_out_label("exported_file.xlsx", database_label("Dummy data store"))
        self.assertEqual(
            exporter.get_icon().exclamation_icon._notifications, ["File extensions don't match the output format."]
        )


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains CustomFileSystemWatcher."""
import os
from PySide6.QtCore import QFileSystemWatcher, Signal, Slot


class CustomFileSystemWatcher(QFileSystemWatcher):
    """A file system watcher that keeps track of renamed files."""

    file_renamed = Signal(str, str)
    file_removed = Signal(str)
    file_added = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._watched_files = {}
        self._watched_dirs = set()
        self._directory_snapshots = {}
        self.directoryChanged.connect(self._handle_dir_changed)

    @Slot(str)
    def _handle_dir_changed(self, dirname):
        """Adds, removes or renames watched files and emits corresponding signals.

        Args:
            dirname (str): path to changed directory
        """
        is_watched_dir = dirname in self._watched_dirs
        watched_files = self._watched_files.setdefault(dirname, set())
        if not is_watched_dir and not watched_files:
            raise RuntimeError("Watching directory that should not be watched.")
        snapshot = self._directory_snapshots[dirname]
        new_snapshot = self._directory_snapshots[dirname] = self._take_snapshot(dirname)
        removed_paths = snapshot - new_snapshot
        added_paths = new_snapshot - snapshot
        for old_path, new_path in zip(list(removed_paths), list(added_paths)):
            # Assume files were renamed.
            # If there was more than one rename, we choose the new name blindly but
            # there doesn't seem to be a better way to do it.
            # Maybe it doesn't even matter? In any case, usually there is just a single rename and nothing else.
            removed_paths.remove(old_path)
            added_paths.remove(new_path)
            if is_watched_dir:
                self.file_renamed.emit(old_path, new_path)
            elif old_path in watched_files:
                watched_files.remove(old_path)
                watched_files.add(new_path)
                self.file_renamed.emit(old_path, new_path)
        for path in removed_paths:
            if is_watched_dir:
                self.file_removed.emit(path)
            elif path in watched_files:
                watched_files.remove(path)
                self.file_removed.emit(path)
        if is_watched_dir:
            for path in added_paths:
                self.file_added.emit(path)
        if not watched_files:
            del self._watched_files[dirname]
            if dirname not in self._watched_dirs:
                self.removePath(dirname)

    def add_persistent_file_path(self, path):
        if not os.path.isfile(path):
            return False
        dirname = os.path.dirname(path)
        self._watched_files.setdefault(dirname, set()).add(path)
        if dirname not in self._directory_snapshots:
            self._directory_snapshots.setdefault(dirname, self._take_snapshot(dirname))
        return self.addPath(dirname)

    def add_persistent_file_paths(self, paths):
        for path in paths:
            self.add_persistent_file_path(path)

    def remove_persistent_file_path(self, path):
        dirname = os.path.dirname(path)
        watched_files = self._watched_files.get(dirname)
        if not watched_files:
            return False
        watched_files.discard(path)
        if watched_files:
            return True
        del self._watched_files[dirname]
        if dirname in self._watched_dirs:
            return True
        del self._directory_snapshots[dirname]
        return self.removePath(dirname)

    def remove_persistent_file_paths(self, paths):
        removed_paths = []
        for path in paths:
            if self.remove_persistent_file_path(path):
                removed_paths.append(path)
        return removed_paths

    def add_persistent_dir_path(self, path):
        if not os.path.isdir(path):
            return False
        self._watched_dirs.add(path)
        self._directory_snapshots.setdefault(path, self._take_snapshot(path))
        return self.addPath(path)

    def remove_persistent_dir_path(self, path):
        self._watched_dirs.discard(path)
        if path in self._watched_files:
            return True
        del self._directory_snapshots[path]
        return self.removePath(path)

    def tear_down(self):
        directories = self.directories()
        if directories:
            self.removePaths(directories)
        self.deleteLater()

    def _take_snapshot(self, dirname):
        return set(self._absfilepaths(dirname))

    @staticmethod
    def _absfilepaths(dirname):
        with os.scandir(dirname) as scan_iterator:
            for entry in scan_iterator:
                if entry.is_file():
                    yield entry.path



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``specification_editor_window`` module."""
import json
import pathlib
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtWidgets import QApplication
from spine_items.exporter.specification import MappingSpecification, MappingType, OutputFormat, Specification
from spine_items.exporter.widgets.specification_editor_window import SpecificationEditorWindow
from spinedb_api.export_mapping.export_mapping import EntityClassMapping, FixedValueMapping
from spinedb_api.export_mapping.export_mapping import from_dict as mappings_from_dict
from spinedb_api.mapping import Position, unflatten
from tests.mock_helpers import clean_up_toolbox, create_toolboxui_with_project


class TestSpecificationEditorWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._temp_dir = TemporaryDirectory()
        self._toolbox = create_toolboxui_with_project(self._temp_dir.name)

    def tearDown(self):
        clean_up_toolbox(self._toolbox)
        self._temp_dir.cleanup()

    def test_empty_editor(self):
        editor = SpecificationEditorWindow(self._toolbox)
        self.assertEqual(editor._ui.mappings_table.model().rowCount(), 1)
        self.assertEqual(editor._ui.mappings_table.model().index(0, 0).data(), "Mapping (1)")
        editor.tear_down()

    def test_mapping_in_table_name_position_disables_fixed_table_name_widgets(self):
        editor = SpecificationEditorWindow(self._toolbox)
        self.assertTrue(editor._ui.fix_table_name_check_box.isEnabled())
        self.assertFalse(editor._ui.fix_table_name_check_box.isChecked())
        self.assertFalse(editor._ui.fix_table_name_line_edit.isEnabled())
        self.assertEqual(editor._ui.fix_table_name_line_edit.text(), "")
        editor._ui.mappings_table.setCurrentIndex(editor._ui.mappings_table.model().index(0, 0))
        model = editor._ui.mapping_table_view.model()
        model.setData(model.index(0, 1), "table name")
        self.assertFalse(editor._ui.fix_table_name_check_box.isEnabled())
        self.assertFalse(editor._ui.fix_table_name_check_box.isChecked())
        self.assertFalse(editor._ui.fix_table_name_line_edit.isEnabled())
        self.assertEqual(editor._ui.fix_table_name_line_edit.text(), "")
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase.tear_down"
        ) as tear_down_window:
            tear_down_window.return_value = True
            editor.tear_down()

    def test_mapping_with_fixed_table_enables_the_check_box_and_fills_the_table_name_field(self):
        flattened_mappings = [FixedValueMapping(Position.table_name, "nice table name"), EntityClassMapping(0)]
        mapping_specification = MappingSpecification(
            MappingType.entities, True, True, "", True, unflatten(flattened_mappings)
        )
        specification = Specification("spec name", mapping_specifications={"my mappings": mapping_specification})
        editor = SpecificationEditorWindow(self._toolbox, specification)
        self.assertTrue(editor._ui.fix_table_name_check_box.isEnabled())
        self.assertTrue(editor._ui.fix_table_name_check_box.isChecked())
        self.assertTrue(editor._ui.fix_table_name_line_edit.isEnabled())
        self.assertEqual(editor._ui.fix_table_name_line_edit.text(), "nice table name")
        editor.tear_down()

    def test_duplicate_specification(self):
        flattened_mappings = [FixedValueMapping(Position.table_name, "nice table name"), EntityClassMapping(0)]
        mapping_specification = MappingSpecification(
            MappingType.entities, True, True, "", True, unflatten(flattened_mappings)
        )
        specification = Specification("spec name", mapping_specifications={"my mappings": mapping_specification})
        editor = SpecificationEditorWindow(self._toolbox, specification)
        editor._spec_toolbar._line_edit_name.setText("my spec name")
        editor._spec_toolbar._line_edit_description.setText("A cool exporter.")
        editor._ui.export_format_combo_box.setCurrentText(OutputFormat.SQL.value)
        with mock.patch.object(self._toolbox, "show_specification_form") as show_duplicate:
            editor._spec_toolbar.duplicate_action.trigger()
            show_duplicate.assert_called_once()
            self.assertEqual(show_duplicate.call_args.args[0], "Exporter")
            self.assertEqual(show_duplicate.call_args.args[1].name, "")
            self.assertEqual(show_duplicate.call_args.args[1].description, "A cool exporter.")
            self.assertEqual(show_duplicate.call_args.args[1].output_format, OutputFormat.SQL)
            self.assertEqual(show_duplicate.call_args.args[1].definition_file_path, "")
            duplicate_mapping_specification_dicts = {
                key: s.to_dict() for key, s in show_duplicate.call_args.args[1].mapping_specifications().items()
            }
            self.assertEqual(duplicate_mapping_specification_dicts, {"my mappings": mapping_specification.to_dict()})
            self.assertIsNot(
                show_duplicate.call_args.args[1].mapping_specifications()["my mappings"], mapping_specification
            )
            self.assertEqual(show_duplicate.call_args.kwargs, {})
        with mock.patch(
            "spinetoolbox.project_item.specification_editor_window.SpecificationEditorWindowBase.tear_down"
        ) as tear_down_window:
            tear_down_window.return_value = True
            editor.tear_down()

    def test_forced_decrease_of_selected_dimension_by_entity_dimensions_is_stored_properly(self):
        mapping_dicts = [
            {"map_type": "EntityClass", "position": 0, "highlight_position": 1},
            {"map_type": "Dimension", "position": "hidden"},
            {"map_type": "Dimension", "position": "hidden"},
            {"map_type": "ParameterDefinition", "position": 3},
            {"map_type": "ParameterValueList", "position": "hidden", "ignorable": True},
            {"map_type": "Entity", "position": "hidden"},
            {"map_type": "Element", "position": "hidden"},
            {"map_type": "Element", "position": "hidden"},
            {"map_type": "Alternative", "position": 4},
            {"map_type": "ParameterValueType", "position": "hidden"},
            {"map_type": "ParameterValue", "position": 5},
        ]
        unflattened_mappings = mappings_from_dict(mapping_dicts)
        mapping_specification = MappingSpecification(
            MappingType.entity_dimension_parameter_values, True, True, "", True, unflattened_mappings
        )
        specification = Specification("spec name", mapping_specifications={"my mappings": mapping_specification})
        specification_path = pathlib.Path(self._temp_dir.name) / "my spec.json"
        specification.definition_file_path = str(specification_path)
        self._toolbox.project().add_specification(specification, save_to_disk=False)
        editor = SpecificationEditorWindow(self._toolbox, specification)
        self.assertEqual(editor._ui.highlight_dimension_spin_box.value(), 2)
        self.assertEqual(editor._ui.entity_dimensions_spin_box.value(), 2)
        editor._spec_toolbar._line_edit_name.setText("my spec name")
        editor._spec_toolbar._line_edit_name.editingFinished.emit()
        editor._ui.entity_dimensions_spin_box.setValue(1)
        self.assertEqual(editor._ui.entity_dimensions_spin_box.value(), 1)
        self.assertEqual(editor._ui.highlight_dimension_spin_box.value(), 1)
        editor.spec_toolbar().save_action.trigger()
        with open(specification_path) as specification_file:
            loaded_specification = Specification.from_dict(json.load(specification_file))
        self.assertEqual(loaded_specification.name, "my spec name")
        expected_dicts = [
            {"map_type": "EntityClass", "position": 0, "highlight_position": 0},
            {"map_type": "Dimension", "position": "hidden"},
            {"map_type": "ParameterDefinition", "position": 3},
            {"map_type": "ParameterValueList", "position": "hidden", "ignorable": True},
            {"map_type": "Entity", "position": "hidden"},
            {"map_type": "Element", "position": "hidden"},
            {"map_type": "Alternative", "position": 4},
            {"map_type": "ParameterValueType", "position": "hidden"},
            {"map_type": "ParameterValue", "position": 5},
        ]
        self.assertEqual(loaded_specification.mapping_specifications()["my mappings"].to_dict()["root"], expected_dicts)
        editor.tear_down()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import argparse
import os.path
from append_license import append_license


def find_ui_dirs(path, found_dirs=None):
    """Recursively searches for 'ui' directories and returns their paths as a list."""
    if found_dirs is None:
        found_dirs = []
    for entry in os.scandir(path):
        if entry.is_dir():
            if entry.name == "ui":
                found_dirs.append(entry.path)
            else:
                find_ui_dirs(entry.path, found_dirs)
    return found_dirs


def fix_resources_imports(path):
    """Fixes resources imports in a given automatically generated Python ui file."""
    lines = []
    with open(path, "r") as in_file:
        for line in in_file:
            if line == "from . import resources_icons_rc\n":
                lines.append("from spine_items import resources_icons_rc\n")
            elif line == "from . import resources_logos_rc\n":
                lines.append("from spine_items import resources_logos_rc\n")
            else:
                lines.append(line)
    with open(path, "w") as out_file:
        out_file.writelines(lines)


def build_ui(input_path, output_path, force):
    """Converts given .ui file to .py."""
    print("Building " + os.path.basename(output_path))
    status = os.system(f"pyside6-uic --from-imports \"{input_path}\" -o \"{output_path}\"")
    if status != 0:
        print("Stop. Build failed.")
        exit(1)
    append_license(output_path)
    fix_resources_imports(output_path)
    append_license(input_path)


def build_qrc(input_path, output_path, force):
    """Converts given .qrc file to .py."""
    print("Building " + os.path.basename(output_path))
    status = os.system(f"pyside6-rcc -o \"{output_path}\" \"{input_path}\"")
    if status != 0:
        print("Stop. Build failed.")
        exit(1)
    append_license(output_path)


print(
    """<Script for Building Spine Items GUI>
Copyright (C) <2017-2022>  <Spine project consortium>
This program comes with ABSOLUTELY NO WARRANTY; for details see 'about'
box in the application. This is free software, and you are welcome to
redistribute it under certain conditions; See files COPYING and
COPYING.LESSER for details."""
)
parser = argparse.ArgumentParser()
parser.add_argument(
    "--force", help="force building of all .ui files", action="store_true"
)
args = parser.parse_args()
script_dir = os.path.dirname(os.path.realpath(__file__))
project_source_dir = os.path.join(script_dir, os.path.pardir, "spine_items")
ui_dirs = find_ui_dirs(project_source_dir)
for ui_dir in ui_dirs:
    print(f"--- Entering {os.path.abspath(ui_dir)} ---")
    ui_entries = []
    py_entries = {}
    for entry in os.scandir(ui_dir):
        if entry.is_file():
            base, extension = os.path.splitext(entry.name)
            if extension == ".ui":
                ui_entries.append(entry)
            elif extension == ".py":
                py_entries[base] = entry
    for ui_entry in ui_entries:
        base, _ = os.path.splitext(ui_entry.name)
        py_entry = py_entries.get(base)
        if py_entry is None or args.force:
            output_name = base + ".py"
            output_path = os.path.join(ui_dir, output_name)
            build_ui(ui_entry.path, output_path, args.force)
            continue
        ui_modification_time = ui_entry.stat().st_mtime
        py_modification_time = py_entry.stat().st_mtime
        if ui_modification_time > py_modification_time:
            build_ui(ui_entry.path, py_entry.path, args.force)
resources_dir = os.path.join(project_source_dir, "ui", "resources")
qrc_entries = []
py_paths = {}
for entry in os.scandir(resources_dir):
    if entry.is_file():
        base, extension = os.path.splitext(entry.name)
        if extension == ".qrc":
            qrc_entries.append(entry)
            output_name = base + "_rc.py"
            output_path = os.path.join(project_source_dir, output_name)
            py_paths[base] = output_path
for qrc_entry in qrc_entries:
    base, _ = os.path.splitext(qrc_entry.name)
    py_path = py_paths.get(base)
    if py_path is None or not os.path.isfile(py_path) or args.force:
        output_name = base + "_rc.py"
        output_path = os.path.join(project_source_dir, output_name)
        build_qrc(qrc_entry.path, output_path, args.force)
        continue
    qrc_modification_time = qrc_entry.stat().st_mtime
    py_modification_time = os.path.getmtime(py_path)
    if qrc_modification_time > py_modification_time:
        build_qrc(qrc_entry.path, py_path, args.force)

print("--- Build completed ---")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for the ``import_sources`` module."""
import unittest
from unittest import mock
from PySide6.QtCore import QModelIndex
from PySide6.QtGui import QUndoStack
from PySide6.QtWidgets import QApplication, QListView, QListWidget, QWidget
from spine_items.importer.connection_manager import ConnectionManager
from spine_items.importer.mvcmodels.mappings_model import MappingsModel
from spine_items.importer.mvcmodels.mappings_model_roles import Role
from spine_items.importer.widgets.import_sources import ImportSources
from spine_items.importer.widgets.table_view_with_button_header import TableViewWithButtonHeader
from spinedb_api.spine_io.importers.reader import Reader, TableProperties
from spinetoolbox.helpers import signal_waiter
from tests.importer.helpers import append_source_table_with_mappings


class TestImportSources(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def setUp(self):
        self._parent_widget = _MockSpecificationEditor()
        self._undo_stack = QUndoStack(self._parent_widget)
        self._mappings_model = MappingsModel(self._undo_stack, self._parent_widget)

    def tearDown(self):
        self._parent_widget.deleteLater()

    def test_connector_fetches_empty_data(self):
        ui = mock.MagicMock()
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        source_list_selection_model = mock.MagicMock()
        ui.source_list.selectionModel.return_value = source_list_selection_model
        source_list_index = mock.MagicMock()
        source_list_index.row.return_value = 0
        source_list_selection_model.currentIndex.return_value = source_list_index
        ui.source_list.selectionModel.return_value = source_list_selection_model
        ui.mapping_list = QListWidget(self._parent_widget)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        data = []
        connection_settings = {"data": {"test data table": data}}
        connector = ConnectionManager(_FixedTableReader, connection_settings, self._parent_widget)
        mapping = {"table_options": {"test data table": {"has_header": False}}}
        import_sources.set_connector(connector, mapping)
        self._mappings_model.append_new_table_with_mapping("test data table", None)
        table_index = self._mappings_model.index(1, 0)
        import_sources._change_selected_table(table_index, QModelIndex())
        source_table_model = import_sources._source_data_model
        with signal_waiter(connector.connection_ready) as waiter:
            connector.init_connection("no file test source")
            waiter.wait()
        with signal_waiter(connector.data_ready) as waiter:
            source_table_model.fetchMore(QModelIndex())
            waiter.wait()
        header = []
        for column in range(source_table_model.columnCount()):
            header.append(source_table_model.headerData(column))
        self.assertEqual(header, [])
        data_rows = []
        for row in range(source_table_model.rowCount()):
            row_data = []
            for column in range(source_table_model.columnCount()):
                row_data.append(source_table_model.index(row, column).data())
            data_rows.append(row_data)
        self.assertEqual(data_rows, [])
        connector.close_connection()

    def test_connector_fetches_data_only(self):
        ui = mock.MagicMock()
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        source_list_selection_model = mock.MagicMock()
        ui.source_list.selectionModel.return_value = source_list_selection_model
        source_list_index = mock.MagicMock()
        source_list_index.row.return_value = 0
        source_list_selection_model.currentIndex.return_value = source_list_index
        ui.source_list.selectionModel.return_value = source_list_selection_model
        ui.mapping_list = QListWidget(self._parent_widget)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        data = [["data 1", "data 2"]]
        connection_settings = {"data": {"test data table": data}}
        connector = ConnectionManager(_FixedTableReader, connection_settings, self._parent_widget)
        mapping = {"table_options": {"test data table": {"has_header": False}}}
        import_sources.set_connector(connector, mapping)
        self._mappings_model.append_new_table_with_mapping("test data table", None)
        table_index = self._mappings_model.index(1, 0)
        import_sources._change_selected_table(table_index, QModelIndex())
        source_table_model = import_sources._source_data_model
        with signal_waiter(connector.connection_ready) as waiter:
            connector.init_connection("no file test source")
            waiter.wait()
        with signal_waiter(connector.data_ready) as waiter:
            source_table_model.fetchMore(QModelIndex())
            waiter.wait()
        header = []
        for column in range(source_table_model.columnCount()):
            header.append(source_table_model.headerData(column))
        self.assertEqual(header, [1, 2])
        data_rows = []
        for row in range(source_table_model.rowCount()):
            row_data = []
            for column in range(source_table_model.columnCount()):
                row_data.append(source_table_model.index(row, column).data())
            data_rows.append(row_data)
        self.assertEqual(data_rows, [["data 1", "data 2"]])
        connector.close_connection()

    def test_connector_fetches_header_only(self):
        ui = mock.MagicMock()
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        source_list_selection_model = mock.MagicMock()
        ui.source_list.selectionModel.return_value = source_list_selection_model
        source_list_index = mock.MagicMock()
        source_list_index.row.return_value = 0
        source_list_selection_model.currentIndex.return_value = source_list_index
        ui.source_list.selectionModel.return_value = source_list_selection_model
        ui.mapping_list = QListWidget(self._parent_widget)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        data = [["header 1", "header 2"]]
        connection_settings = {"data": {"test data table": data}}
        connector = ConnectionManager(_FixedTableReader, connection_settings, self._parent_widget)
        mapping = {"table_options": {"test data table": {"has_header": True}}}
        import_sources.set_connector(connector, mapping)
        self._mappings_model.append_new_table_with_mapping("test data table", None)
        table_index = self._mappings_model.index(1, 0)
        import_sources._change_selected_table(table_index, QModelIndex())
        source_table_model = import_sources._source_data_model
        with signal_waiter(connector.connection_ready) as waiter:
            connector.init_connection("no file test source")
            waiter.wait()
        with signal_waiter(connector.data_ready) as waiter:
            source_table_model.fetchMore(QModelIndex())
            waiter.wait()
        header = []
        for column in range(source_table_model.columnCount()):
            header.append(source_table_model.headerData(column))
        self.assertEqual(header, ["header 1", "header 2"])
        data_rows = []
        for row in range(source_table_model.rowCount()):
            row_data = []
            for column in range(source_table_model.columnCount()):
                row_data.append(source_table_model.index(row, column).data())
            data_rows.append(row_data)
        self.assertEqual(data_rows, [])
        connector.close_connection()

    def test_connector_fetches_data_and_header(self):
        ui = mock.MagicMock()
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        source_list_selection_model = mock.MagicMock()
        ui.source_list.selectionModel.return_value = source_list_selection_model
        source_list_index = mock.MagicMock()
        source_list_index.row.return_value = 0
        source_list_selection_model.currentIndex.return_value = source_list_index
        ui.source_list.selectionModel.return_value = source_list_selection_model
        ui.mapping_list = QListView(self._parent_widget)
        ui.mapping_list.setModel(self._mappings_model)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        data = [["header 1", "header 2"], ["data 1", "data 2"]]
        connection_settings = {"data": {"test data table": data}}
        connector = ConnectionManager(_FixedTableReader, connection_settings, self._parent_widget)
        mapping = {"table_options": {"test data table": {"has_header": True}}}
        import_sources.set_connector(connector, mapping)
        self._mappings_model.append_new_table_with_mapping("test data table", None)
        table_index = self._mappings_model.index(1, 0)
        import_sources._change_selected_table(table_index, QModelIndex())
        source_table_model = import_sources._source_data_model
        with signal_waiter(connector.connection_ready) as waiter:
            connector.init_connection("no file test source")
            waiter.wait()
        with signal_waiter(connector.data_ready) as waiter:
            source_table_model.fetchMore(QModelIndex())
            waiter.wait()
        header = []
        for column in range(source_table_model.columnCount()):
            header.append(source_table_model.headerData(column))
        self.assertEqual(header, ["header 1", "header 2"])
        data_rows = []
        for row in range(source_table_model.rowCount()):
            row_data = []
            for column in range(source_table_model.columnCount()):
                row_data.append(source_table_model.index(row, column).data())
            data_rows.append(row_data)
        self.assertEqual(data_rows, [["data 1", "data 2"]])
        connector.close_connection()

    def test_copying_and_pasting_mappings_and_options_from_one_table_to_another(self):
        ui = mock.MagicMock()
        ui.source_list = QListView(self._parent_widget)
        ui.source_list.setModel(self._mappings_model)
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        ui.mapping_list = QListView(self._parent_widget)
        ui.mapping_list.setModel(self._mappings_model)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        data = [["header 1", "header 2"], ["data 1", "data 2"]]
        connection_settings = {
            "data": {"source table": data, "destination table": data},
            "options": {"source table": {"has_header": True}, "destination table": {"has_header": False}},
        }
        connector = ConnectionManager(_FixedTableReader, connection_settings, self._parent_widget)
        mapping = {}
        import_sources.set_connector(connector, mapping)
        with signal_waiter(connector.connection_ready) as waiter:
            connector.init_connection("no file test source")
            waiter.wait()
        self.assertTrue(connector.is_connected)
        connector.request_tables()
        while self._mappings_model.rowCount() != 3:
            QApplication.processEvents()
        source_table_rows = {}
        for row in range(1, self._mappings_model.rowCount()):
            source_table_rows[self._mappings_model.index(row, 0).data(Role.ITEM).name] = row
        self.assertIn("source table", source_table_rows)
        source_table_index = self._mappings_model.index(source_table_rows["source table"], 0)
        ui.source_list.setCurrentIndex(source_table_index)
        with mock.patch("spine_items.importer.widgets.import_sources.QApplication") as mock_application:
            clipboard = _MockClipboard()
            mock_application.clipboard.return_value = clipboard
            import_sources._copy_mappings_and_options_to_clipboard(source_table_index)
            self.assertIn("destination table", source_table_rows)
            destination_table_index = self._mappings_model.index(source_table_rows["destination table"], 0)
            ui.source_list.setCurrentIndex(destination_table_index)
            self.assertEqual(
                connector.table_options,
                {"source table": {"has_header": True}, "destination table": {"has_header": False}},
            )
            import_sources._paste_mapping_list_from_clipboard()
            self.assertEqual(self._mappings_model.rowCount(source_table_index), 1)
            self.assertEqual(self._mappings_model.rowCount(destination_table_index), 1)
            self.assertEqual(self._mappings_model.index(0, 0, destination_table_index).data(), "Mapping 1")
            self.assertEqual(
                connector.table_options,
                {"source table": {"has_header": True}, "destination table": {"has_header": False}},
            )
            import_sources._paste_table_options_from_clipboard()
        self.assertEqual(
            connector.table_options, {"source table": {"has_header": True}, "destination table": {"has_header": True}}
        )
        connector.close_connection()

    def test_copying_and_pasting_mappings_from_one_table_to_another(self):
        ui = mock.MagicMock()
        ui.source_list = QListView(self._parent_widget)
        ui.source_list.setModel(self._mappings_model)
        ui.source_data_table = TableViewWithButtonHeader(self._parent_widget)
        ui.mapping_list = QListView(self._parent_widget)
        ui.mapping_list.setModel(self._mappings_model)
        import_sources = ImportSources(self._mappings_model, ui, self._undo_stack, self._parent_widget)
        source_table_index = append_source_table_with_mappings(
            self, self._mappings_model, "source table", "one mapping", []
        )
        destination_table_index = append_source_table_with_mappings(
            self, self._mappings_model, "destination table", "another mapping", []
        )
        with mock.patch("spine_items.importer.widgets.import_sources.QApplication") as mock_application:
            clipboard = _MockClipboard()
            mock_application.clipboard.return_value = clipboard
            import_sources._copy_mapping_list_to_clipboard(source_table_index)
            ui.source_list.setCurrentIndex(destination_table_index)
            import_sources._paste_mapping_list_from_clipboard()
        self.assertEqual(self._mappings_model.rowCount(source_table_index), 1)
        self.assertEqual(self._mappings_model.rowCount(destination_table_index), 1)
        self.assertEqual(self._mappings_model.index(0, 0, destination_table_index).data(), "one mapping")


class _MockSpecificationEditor(QWidget):
    def is_file_less(self):
        return False


class _FixedTableReader(Reader):
    DISPLAY_NAME = "test data source"
    OPTIONS = {"has_header": {"type": bool, "label": "Has header", "default": False}}
    FILE_EXTENSIONS = {"*.*"}

    def __init__(self, settings):
        super().__init__(None)
        self._data = settings["data"]
        self._options = settings.get("options")

    def connect_to_source(self, source, **extras):
        pass

    def disconnect(self):
        pass

    def get_tables_and_properties(self):
        return {table: TableProperties(self._options.get(table, {})) for table in self._data}

    def get_data_iterator(self, table, options, max_rows=-1):
        if options["has_header"]:
            header = self._data[table][0]
            begin = 1
            end = (begin + max_rows) if max_rows > -1 else None
        else:
            header = []
            begin = 0
            end = max_rows if max_rows > -1 else None
        return self._data[table][slice(begin, end)], header


class _MockClipboard:
    mime_data = None

    def setMimeData(self, data):
        self.mime_data = data

    def mimeData(self):
        return self.mime_data


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Utilities to validate that a database exists."""
from pathlib import Path
from PySide6.QtCore import QObject, QRunnable, QThread, QThreadPool, QTimer, Signal, Slot
from PySide6.QtWidgets import QApplication
from spine_items.utils import check_database_url


class _ValidationTask(QRunnable):
    """Runnable that performs the actual validation in a worker thread."""

    def __init__(self, dialect, sa_url, finish_slot, fail_slot, success_slot):
        """
        Args:
            dialect (str): database dialect
            sa_url (URL): SQLAlchemy URL
            finish_slot (Callable): a Qt slot to connect to the finished signal
            fail_slot (Callable): a Qt slot to connect to the fail signal
            success_slot (Callable, optional): a Qt slot to connect to the success signal
        """
        super().__init__()
        self._dialect = dialect
        self._sa_url = sa_url
        self._signals = _TaskSignals()
        self._signals.moveToThread(None)
        self._signals.validation_failed.connect(fail_slot)
        if success_slot is not None:
            self._signals.validation_succeeded.connect(success_slot)
        self._signals.finished.connect(finish_slot)

    def run(self):
        """Checks if it is possible to establish an SQLAlchemy connection to the given URL."""
        try:
            self._signals.moveToThread(QThread.currentThread())
            if self._dialect == "sqlite":
                database_path = Path(self._sa_url.database)
                if not database_path.exists():
                    self._signals.validation_failed.emit(
                        "File does not exist. Check the Database field in the URL.", self._sa_url
                    )
                    return
                elif database_path.is_dir():
                    self._signals.validation_failed.emit(
                        "Database points to a directory, not a file. Check the Database field in the URL.",
                        self._sa_url,
                    )
                    return
                elif database_path.stat().st_size == 0:
                    self._signals.validation_failed.emit(
                        "Database points to an empty file. Check the Database field in the URL.",
                        self._sa_url,
                    )
                    return
            error = check_database_url(self._sa_url)
            if error is not None:
                self._signals.validation_failed.emit(error, self._sa_url)
                return
            self._signals.validation_succeeded.emit(self._sa_url)
        finally:
            self._signals.finished.emit()
            application = QApplication.instance()
            self._signals.moveToThread(application.thread())
            self._signals.deleteLater()


class _TaskSignals(QObject):
    """Signals for validation task."""

    validation_failed = Signal(str, object)
    validation_succeeded = Signal(object)
    finished = Signal()


class DatabaseConnectionValidator(QObject):
    """Database connection validator.

    Executes a single validation task at a time in a separate thread.
    While the thread is busy, queues the latest validation request.
    """

    def __init__(self, parent=None):
        """
        Args:
            parent (QObject): parent object
        """
        super().__init__(parent)
        self._threadpool = QThreadPool.globalInstance()
        self._busy = False
        self._deferred_task = None
        self._closed = False

    def is_busy(self):
        """Tests if there is a validator task running."""
        return self._busy

    def validate_url(self, dialect, sa_url, fail_slot, success_slot):
        """Connects signals and starts a task to validate the given URL.

        If validation task is running, puts a new task to a queue.
        If there is a task in the queue, it is replaced by a new one.

        If validation succeeds, success_slot will be called with no arguments.
        If validation fails, fail_slot will be called with an error string parameter.

        Args:
            dialect (str): database dialect
            sa_url (URL): SQLAlchemy URL
            fail_slot (Callable): a Qt slot that is invoked if validation fails
            success_slot (Callable, optional): a Qt Slot that is invoked if validation succeeds
        """
        if self._closed:
            return
        if not self._busy:
            self._busy = True
            self._threadpool.start(_ValidationTask(dialect, sa_url, self._set_non_busy, fail_slot, success_slot))
        else:
            queued = self._deferred_task is not None
            self._deferred_task = _ValidationTask(dialect, sa_url, self._set_non_busy, fail_slot, success_slot)
            if not queued:
                QTimer.singleShot(500, self._start_from_queue)

    def _start_from_queue(self):
        """Tries to start a deferred task from the queue."""
        if self._closed:
            return
        if self._busy:
            QTimer.singleShot(500, self._start_from_queue)
            return
        self._busy = True
        self._threadpool.start(self._deferred_task)
        self._deferred_task = None

    @Slot()
    def _set_non_busy(self):
        """Sets the machine not busy."""
        self._busy = False

    def wait_for_finish(self):
        """Waits for the thread to finish."""
        self._closed = True
        while self._busy:
            QApplication.processEvents()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for :class:`DataTransformerExecutable`."""
from multiprocessing import Lock
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock
from spine_engine.project_item.project_item_resource import database_resource
from spine_engine.utils.helpers import ExecutionDirection
from spine_items.data_transformer.data_transformer_specification import DataTransformerSpecification
from spine_items.data_transformer.executable_item import ExecutableItem
from spine_items.data_transformer.settings import EntityClassRenamingSettings
from spinedb_api import append_filter_config


class TestDataTransformerExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        logger = MagicMock()
        item = ExecutableItem("T", None, self._temp_dir.name, logger)
        self.assertEqual(item.item_type(), "Data Transformer")

    def test_execute(self):
        logger = MagicMock()
        item = ExecutableItem("T", None, self._temp_dir.name, logger)
        db_resource = database_resource("provider", "sqlite:///db.sqlite")
        self.assertTrue(item.execute([db_resource], [], Lock()))
        expected_resource = database_resource(item.name, "sqlite:///db.sqlite")
        self.assertEqual(item.output_resources(ExecutionDirection.FORWARD), [expected_resource])

    def test_skip_execution(self):
        logger = MagicMock()
        item = ExecutableItem("T", None, self._temp_dir.name, logger)
        db_resource = database_resource("provider", "sqlite:///db.sqlite")
        item.exclude_execution([db_resource], [], Lock())
        expected_resource = database_resource(item.name, "sqlite:///db.sqlite")
        self.assertEqual(item.output_resources(ExecutionDirection.FORWARD), [expected_resource])

    def test_execute_with_specification(self):
        settings = EntityClassRenamingSettings({})
        specification = DataTransformerSpecification("specification", settings, "test specification")
        logger = MagicMock()
        transformer = ExecutableItem("T", specification, self._temp_dir.name, logger)
        db_resource = database_resource("provider", "sqlite:///db.sqlite")
        self.assertTrue(transformer.execute([db_resource], [], Lock()))
        filter_url = append_filter_config("sqlite:///db.sqlite", transformer._filter_config_path)
        expected_resource = database_resource(transformer.name, filter_url)
        self.assertEqual(transformer.output_resources(ExecutionDirection.FORWARD), [expected_resource])

    def test_skip_execution_with_specification(self):
        settings = EntityClassRenamingSettings({})
        specification = DataTransformerSpecification("specification", settings, "test specification")
        logger = MagicMock()
        transformer = ExecutableItem("T", specification, self._temp_dir.name, logger)
        db_resource = database_resource("provider", "sqlite:///db.sqlite")
        transformer.exclude_execution([db_resource], [], Lock())
        filter_url = append_filter_config("sqlite:///db.sqlite", transformer._filter_config_path)
        expected_resource = database_resource(transformer.name, filter_url)
        self.assertEqual(transformer.output_resources(ExecutionDirection.FORWARD), [expected_resource])


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""The ToolFactory class."""
import os
import uuid
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QFileDialog
from spinetoolbox.project_item.project_item_factory import ProjectItemFactory
from .tool import Tool
from .tool_icon import ToolIcon
from .widgets.add_tool_widget import AddToolWidget
from .widgets.custom_menus import ToolSpecificationMenu
from .widgets.tool_properties_widget import ToolPropertiesWidget
from .widgets.tool_specification_editor_window import ToolSpecificationEditorWindow


class ToolFactory(ProjectItemFactory):
    @staticmethod
    def item_class():
        return Tool

    @staticmethod
    def icon():
        return ":/icons/item_icons/hammer.svg"

    @staticmethod
    def icon_color():
        return QColor("red")

    @staticmethod
    def make_add_item_widget(toolbox, x, y, specification):
        return AddToolWidget(toolbox, x, y, specification)

    @staticmethod
    def make_icon(toolbox):
        return ToolIcon(toolbox, ToolFactory.icon(), ToolFactory.icon_color())

    @staticmethod
    def make_item(name, item_dict, toolbox, project):
        return Tool.from_dict(name, item_dict, toolbox, project)

    @staticmethod
    def make_properties_widget(toolbox):
        return ToolPropertiesWidget(toolbox)

    @staticmethod
    def make_specification_menu(parent, index):
        return ToolSpecificationMenu(parent, index)

    @staticmethod
    def make_specification_editor(toolbox, specification=None, item=None, **kwargs):
        return ToolSpecificationEditorWindow(toolbox, specification, item, **kwargs)

    @staticmethod
    def repair_specification(toolbox, specification):
        """See base class."""
        ProjectItemFactory.repair_specification(toolbox, specification)
        # Skip for Executable Tool Specs because it is allowed to not have a main program
        if specification.tooltype == "executable":
            return
        # Check that main program file exists. If not, log a message with an anchor to find it
        filename = specification.includes[0]
        full_path = os.path.join(specification.path, filename)
        if not os.path.isfile(full_path):
            url = str(uuid.uuid4())
            toolbox.register_anchor_callback(
                url, lambda t=toolbox, f=filename, s=specification: _find_main_program_file(t, filename, specification)
            )
            toolbox.msg_error.emit(
                f"Tool spec <b>{specification.name}</b> won't work because "
                f"main program file <b>{full_path}</b> doesn't exist. "
                f"<a style='color:white;' href='{url}'><b>[find it]</b></a>"
            )


def _find_main_program_file(toolbox, filename, specification):
    """Shows a file dialog where the user can find the missing main program file of a tool spec.
    Updates the spec if good.

    Args:
        toolbox (ToolboxUI): The toolbox QMainWindow
        filename (str): The name of the main program file
        specification (ToolSpecification): The spec to update
    """
    answer = QFileDialog.getOpenFileName(toolbox, f"Select {filename}", toolbox.project().project_dir)
    if answer[0] == "":  # Cancel button clicked
        return
    specification = specification.clone()
    specification.path, specification.includes[0] = os.path.split(answer[0])
    toolbox.project().replace_specification(specification.name, specification)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for MergerExecutable."""
import gc
from multiprocessing import Lock
import os.path
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine.project_item.connection import Connection
from spine_engine.project_item.project_item_resource import database_resource
from spine_engine.spine_engine import SpineEngine, SpineEngineState
from spine_items.merger.executable_item import ExecutableItem
from spinedb_api import DatabaseMapping, create_new_spine_database, import_functions
from spinedb_api.spine_db_server import db_server_manager


class TestMergerExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        gc.collect()
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "Merger")

    def test_from_dict(self):
        name = "Output Data Store"
        item_dict = {"type": "Data Store", "description": "", "x": 0, "y": 0, "cancel_on_error": True}
        logger = mock.MagicMock()
        item = ExecutableItem.from_dict(item_dict, name, self._temp_dir.name, None, {}, logger)
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("Merger", item.item_type())

    def test_stop_execution(self):
        executable = ExecutableItem("name", True, self._temp_dir.name, mock.MagicMock())
        with mock.patch(
            "spine_engine.project_item.executable_item_base.ExecutableItemBase.stop_execution"
        ) as mock_stop_execution:
            executable.stop_execution()
            mock_stop_execution.assert_called_once()

    def test_execute(self):
        executable = ExecutableItem("name", True, self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))

    def test_execute_merge_two_dbs(self):
        """Creates two db's with some data and merges them to a third db."""
        db1_path = Path(self._temp_dir.name, "db1.sqlite")
        db1_url = "sqlite:///" + str(db1_path)
        # Add some data to db1
        with DatabaseMapping(db1_url, create=True) as db1_map:
            import_functions.import_entity_classes(db1_map, [("a",)])
            import_functions.import_entities(db1_map, [("a", "a_1")])
            db1_map.commit_session("Add an object class 'a' and an object for unit tests.")
        db1_map.close()
        db2_path = Path(self._temp_dir.name, "db2.sqlite")
        db2_url = "sqlite:///" + str(db2_path)
        # Add some data to db2
        with DatabaseMapping(db2_url, create=True) as db2_map:
            import_functions.import_entity_classes(db2_map, [("b",)])
            import_functions.import_entities(db2_map, [("b", "b_1")])
            db2_map.commit_session("Add an object class 'b' and an object for unit tests.")
        db2_map.close()
        # Make an empty output db
        db3_path = Path(self._temp_dir.name, "db3.sqlite")
        db3_url = "sqlite:///" + str(db3_path)
        create_new_spine_database(db3_url)
        logger = mock.MagicMock()
        logger.__reduce__ = lambda _: (mock.MagicMock, ())
        executable = ExecutableItem("name", True, self._temp_dir.name, logger)
        input_db_resources = [database_resource("provider", db1_url), database_resource("provider", db2_url)]
        output_db_resources = [database_resource("receiver", db3_url)]
        with db_server_manager() as mngr_queue:
            for r in input_db_resources + output_db_resources:
                r.metadata["db_server_manager_queue"] = mngr_queue
            self.assertTrue(executable.execute(input_db_resources, output_db_resources, Lock()))
        # Check output db
        with DatabaseMapping(db3_url) as output_db_map:
            class_list = output_db_map.query(output_db_map.entity_class_sq).all()
            self.assertEqual(len(class_list), 2)
            self.assertEqual(class_list[0].name, "a")
            self.assertEqual(class_list[1].name, "b")
            entity_list_a = output_db_map.query(output_db_map.entity_sq).filter_by(class_id=class_list[0].id).all()
            self.assertEqual(len(entity_list_a), 1)
            self.assertEqual(entity_list_a[0].name, "a_1")
            entity_list_b = output_db_map.query(output_db_map.entity_sq).filter_by(class_id=class_list[1].id).all()
            self.assertEqual(len(entity_list_b), 1)
            self.assertEqual(entity_list_b[0].name, "b_1")
        output_db_map.close()

    def test_write_order(self):
        db1_path = Path(self._temp_dir.name, "db1.sqlite")
        db1_url = "sqlite:///" + str(db1_path)
        # Add some data to db1
        with DatabaseMapping(db1_url, create=True) as db1_map:
            import_functions.import_data(db1_map, entity_classes=[("fish",)])
            db1_map.commit_session("Add test data.")
        db1_map.close()
        db2_path = Path(self._temp_dir.name, "db2.sqlite")
        db2_url = "sqlite:///" + str(db2_path)
        # Add some data to db2
        with DatabaseMapping(db2_url, create=True) as db2_map:
            import_functions.import_data(db2_map, entity_classes=[("cat",)])
            db2_map.commit_session("Add test data.")
        db2_map.close()
        # Make an empty output db
        db3_path = Path(self._temp_dir.name, "db3.sqlite")
        db3_url = "sqlite:///" + str(db3_path)
        # Make two mergers
        logger = mock.MagicMock()
        logger.__reduce__ = lambda _: (mock.MagicMock, ())
        merger1 = {"type": "Merger", "description": "", "cancel_on_error": True}
        merger2 = {"type": "Merger", "description": "", "cancel_on_error": True}
        # Make two input DS and one output DS
        input_ds_1 = {
            "type": "Data Store",
            "description": "",
            "url": {
                "dialect": "sqlite",
                "database": {
                    "type": "path",
                    "relative": False,
                    "path": str(db1_path),
                },
            },
        }
        input_ds_2 = {
            "type": "Data Store",
            "description": "",
            "url": {
                "dialect": "sqlite",
                "database": {
                    "type": "path",
                    "relative": False,
                    "path": str(db2_path),
                },
            },
        }
        output_ds = {
            "type": "Data Store",
            "description": "",
            "url": {
                "dialect": "sqlite",
                "database": {
                    "type": "path",
                    "relative": False,
                    "path": str(db3_path),
                },
            },
        }
        # Make connections
        conn1in = Connection("ds1", "right", "merger1", "left")
        conn2in = Connection("ds2", "right", "merger2", "left")
        conn1out = Connection("merger1", "left", "output_ds", "right", options={"write_index": 2})
        conn2out = Connection("merger2", "left", "output_ds", "right", options={"write_index": 1})
        # Make and run engine
        items = {"ds1": input_ds_1, "ds2": input_ds_2, "output_ds": output_ds, "merger1": merger1, "merger2": merger2}
        execution_permits = {name: True for name in items}
        # We can't easily enforce merger1 to execute before merger2 so the write index matters...
        # So for the moment, let's run 5 times and hope
        for _ in range(5):
            engine = SpineEngine(
                items=items,
                connections=[x.to_dict() for x in (conn1in, conn2in, conn1out, conn2out)],
                execution_permits=execution_permits,
                project_dir=self._temp_dir.name,
            )
            create_new_spine_database(db3_url)
            engine.run()
            self.assertEqual(engine.state(), SpineEngineState.COMPLETED)
            with DatabaseMapping(db3_url, create=True) as db3_map:
                commits = db3_map.query(db3_map.commit_sq).all()
            db3_map.close()
            merger1_idx = next(iter(k for k, commit in enumerate(commits) if db1_url in commit.comment))
            merger2_idx = next(iter(k for k, commit in enumerate(commits) if db2_url in commit.comment))
            self.assertTrue(merger2_idx < merger1_idx)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains ToolInstance classes."""
import os
import sys
from spine_engine.execution_managers.kernel_execution_manager import KernelExecutionManager
from spine_engine.execution_managers.persistent_execution_manager import (
    JuliaPersistentExecutionManager,
    PythonPersistentExecutionManager,
)
from spine_engine.execution_managers.process_execution_manager import ProcessExecutionManager
from spine_engine.utils.helpers import resolve_conda_executable, resolve_gams_executable, resolve_python_interpreter
from spine_items.tool.utils import get_julia_path_and_project
from spine_items.utils import escape_backward_slashes


class ToolInstance:
    """Tool instance base class."""

    def __init__(self, tool_specification, basedir, settings, logger, owner):
        """
        Args:
            tool_specification (ToolSpecification): the tool specification for this instance
            basedir (str): the path to the directory where this instance should run
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): a logger instance
            owner (ExecutableItemBase): The item that owns the instance
        """
        self.tool_specification = tool_specification
        self.basedir = basedir
        self._settings = settings
        self._logger = logger
        self._owner = owner
        self.exec_mngr = None
        self.program = None  # Program to start in the subprocess
        self.args = []  # List of command line arguments for the program
        self.killed = False

    def _update_killed(self):
        try:
            self.killed = self.exec_mngr.killed
        except AttributeError:
            pass

    @property
    def owner(self):
        return self._owner

    def is_running(self):
        return self.exec_mngr is not None

    def terminate_instance(self):
        """Terminates Tool instance execution."""
        if not self.exec_mngr:
            return
        self.exec_mngr.stop_execution()
        self.exec_mngr = None

    def prepare(self, args):
        """Prepares this instance for execution.

        Implement in subclasses to perform specific initialization.

        Args:
            args (list): Tool cmd line args
        """
        raise NotImplementedError()

    def execute(self):
        """Executes a prepared instance. Implement in subclasses."""
        raise NotImplementedError()

    def append_cmdline_args(self, tool_args):
        """
        Appends Tool specification command line args into instance args list.

        Args:
            tool_args (list): List of Tool cmd line args
        """
        self.args += self.tool_specification.cmdline_args
        if tool_args:
            self.args += tool_args


class GAMSToolInstance(ToolInstance):
    """Class for GAMS Tool instances."""

    def prepare(self, args):
        """See base class."""
        gams_path = self._settings.value("appSettings/gamsPath", defaultValue="")
        self.program = resolve_gams_executable(gams_path)
        self.args.append(self.tool_specification.main_prgm)
        self.args.append(f"curDir={self.basedir}")  # Set gams current working directory
        self.args.append("logoption=3")  # This could be an option in Settings
        self.append_cmdline_args(args)
        # Note: workdir=self.basedir sets the working directory for the subprocess.Popen process.
        self.exec_mngr = ProcessExecutionManager(self._logger, self.program, self.args, workdir=self.basedir)

    def execute(self):
        """Executes a prepared instance."""
        ret = self.exec_mngr.run_until_complete()
        if ret != 0:
            try:
                return_msg = self.tool_specification.return_codes[ret]
                self._logger.msg_error.emit(f"<b>{return_msg}</b> [exit code:{ret}]")
            except KeyError:
                self._logger.msg_error.emit(f"Unknown return code ({ret})")
            debug_anchor = self.tool_specification.make_debug_project(self.basedir)
            if debug_anchor is not None:
                self._logger.msg.emit(f"{debug_anchor}")
        self._update_killed()
        self.exec_mngr = None
        return ret


class JuliaToolInstance(ToolInstance):
    """Class for Julia Tool instances."""

    def __init__(self, tool_specification, basedir, settings, kill_completed_processes, logger, owner):
        """
        Args:
            tool_specification (ToolSpecification): the tool specification for this instance
            basedir (str): the path to the directory where this instance should run
            settings (QSettings): Toolbox settings
            kill_completed_processes (bool): whether to kill completed persistent processes
            logger (LoggerInterface): a logger instance
            owner (ExecutableItemBase): The item that owns the instance
        """
        super().__init__(tool_specification, basedir, settings, logger, owner)
        self._kill_completed_processes = kill_completed_processes

    def prepare_sysimg_maker(self, julia_args, cmdline_args):
        """Prepare this instance for creating a sysimage.

        Args:
            julia_args (list): First item is the path to Julia executable, the remaining items
            are the initial args for Julia (ie. '--project=some/path')
            cmdline_args (list): Tool and Tool Spec cmd line args for the Julia script

        Returns:
            bool: True if making a sysimage has been enabled, False otherwise
        """
        make_sysimage = self._settings.value("appSettings/makeSysImage", defaultValue="false")
        if make_sysimage == "true":
            if self.tool_specification.main_prgm:
                julia_args.append(self.tool_specification.main_prgm)
            julia_args += cmdline_args
            self.exec_mngr = ProcessExecutionManager(self._logger, julia_args[0], julia_args[1:], workdir=self.basedir)
            return True
        return False

    def make_julia_commands(self, cmdline_args):
        """Returns a list of commands to execute in the Jupyter or in the Basic Console.
        Given cmdline_args (Tool and Tool Spec args) are embedded into the commands.

        Args:
            cmdline_args (list): Tool and Tool Spec cmd line args for the Julia script

        Returns:
            list of str: Commands to execute in the Julia Console
        """
        mod_work_dir = escape_backward_slashes(self.basedir)
        mod_main_prgm = escape_backward_slashes(self.tool_specification.main_prgm)
        cmds = [f'cd("{mod_work_dir}");']
        if cmdline_args:
            fmt_cmdline_args = '["' + escape_backward_slashes('", "'.join(cmdline_args)) + '"]'
            cmds += [f"empty!(ARGS); append!(ARGS, {fmt_cmdline_args});"]
        cmds += [f'include("{mod_main_prgm}")']
        return cmds

    def make_sysimage_arg(self):
        """Returns a '--sysimage=path/to/sysimage' arg for the Julia
        process or None if the sysimage path is missing or invalid."""
        sysimage_path = self._owner.options.get("julia_sysimage", "")
        if sysimage_path == "":
            return None
        if sysimage_path != "" and not os.path.isfile(sysimage_path):
            self._logger.msg_error.emit(f"Ignoring Sysimage <b>{sysimage_path}</b> because it does not exist")
            return None
        return f"--sysimage={sysimage_path}"

    def prepare(self, args):
        """See base class."""
        self.tool_specification.init_execution_settings()  # Set default execution settings if they are missing
        julia_args = get_julia_path_and_project(self.tool_specification.execution_settings, self._settings)
        if not julia_args:
            k_name = self.tool_specification.execution_settings["kernel_spec_name"]
            self._logger.msg_error.emit(f"Invalid kernel '{k_name}'. Missing resource dir or corrupted kernel.json.")
            raise RuntimeError()
        cmdline_args = self.tool_specification.cmdline_args + args
        if self.prepare_sysimg_maker(julia_args, cmdline_args):
            return
        sysimage_arg = self.make_sysimage_arg()
        commands = self.make_julia_commands(cmdline_args)
        if self.tool_specification.execution_settings["use_jupyter_console"]:
            server_ip = "127.0.0.1"
            if self._settings.value("engineSettings/remoteExecutionEnabled", defaultValue="false") == "true":
                server_ip = self._settings.value("engineSettings/remoteHost", "")
            kernel_name = self.tool_specification.execution_settings["kernel_spec_name"]
            extra_switches = [sysimage_arg] if sysimage_arg else None
            self.exec_mngr = KernelExecutionManager(
                self._logger,
                kernel_name,
                commands,
                kill_completed=self._kill_completed_processes,
                group_id=self.owner.group_id,
                extra_switches=extra_switches,
                server_ip=server_ip,
            )
            return
        if sysimage_arg:
            julia_args.append(sysimage_arg)
        alias = f"julia {' '.join([self.tool_specification.main_prgm, *cmdline_args])}"
        self.exec_mngr = JuliaPersistentExecutionManager(
            self._logger, julia_args, commands, alias, self._kill_completed_processes, self.owner.group_id
        )

    def execute(self):
        """Executes a prepared instance."""
        ret = self.exec_mngr.run_until_complete()
        if ret != 0:
            try:
                return_msg = self.tool_specification.return_codes[ret]
                self._logger.msg_error.emit(f"\t<b>{return_msg}</b> [exit code:{ret}]")
            except KeyError:
                self._logger.msg_error.emit(f"\tUnknown return code ({ret})")
        self._update_killed()
        self.exec_mngr = None
        return ret


class PythonToolInstance(ToolInstance):
    """Class for Python Tool instances."""

    def __init__(self, tool_specification, basedir, settings, kill_completed_processes, logger, owner):
        """
        Args:
            tool_specification (ToolSpecification): the tool specification for this instance
            basedir (str): the path to the directory where this instance should run
            settings (QSettings): Toolbox settings
            kill_completed_processes (bool): whether to kill completed persistent processes
            logger (LoggerInterface): a logger instance
            owner (ExecutableItemBase): The item that owns the instance
        """
        super().__init__(tool_specification, basedir, settings, logger, owner)
        self._kill_completed_processes = kill_completed_processes

    def make_python_jupyter_console_commands(self, cmdline_args):
        """Returns commands to execute in the Python Jupyter Console.

        Args:
            cmdline_args (list): Tool and Tool Spec cmd line args

        Returns:
            list: List of commands for the Python Jupyter Console
        """
        cd_command = f"%cd -q {self.basedir}"  # -q: quiet
        main_command = f'%run "{self.tool_specification.main_prgm}"'
        if cmdline_args:
            main_command += " " + '"' + '" "'.join(cmdline_args) + '"'
        return [cd_command, main_command]

    def make_python_basic_console_commands(self, cmdline_args):
        """Returns commands to execute in the Python Basic Console.

        Args:
            cmdline_args (list): Tool and Tool Spec cmd line args

        Returns:
            list: List of commands for the Python Basic Console
        """
        commands = []
        fp = self.tool_specification.main_prgm.replace(os.sep, "/")
        full_fp = os.path.join(self.basedir, self.tool_specification.main_prgm).replace(os.sep, "/")
        commandline_args = [full_fp] + cmdline_args
        fmt_cmdline_args = '["' + escape_backward_slashes('", "'.join(commandline_args)) + '"]'
        commands.append(f"import sys; sys.argv = {fmt_cmdline_args};")
        basedir = self.basedir.replace(os.sep, "/")
        commands.append(f"import os; os.chdir({repr(basedir)})")
        commands += self._make_exec_code(fp, full_fp)
        return commands

    def prepare(self, args):
        """See base class."""
        self.tool_specification.init_execution_settings()  # Initialize execution settings
        cmdline_args = self.tool_specification.cmdline_args + args
        if self.tool_specification.execution_settings["use_jupyter_console"]:
            server_ip = "127.0.0.1"
            if self._settings.value("engineSettings/remoteExecutionEnabled", defaultValue="false") == "true":
                server_ip = self._settings.value("engineSettings/remoteHost", "")
            kernel_name = self.tool_specification.execution_settings["kernel_spec_name"]
            commands = self.make_python_jupyter_console_commands(cmdline_args)
            env = self.tool_specification.execution_settings["env"]  # Activate environment if "conda"
            conda_exe = resolve_conda_executable(self._settings.value("appSettings/condaPath", defaultValue=""))
            self.exec_mngr = KernelExecutionManager(
                self._logger,
                kernel_name,
                commands,
                kill_completed=self._kill_completed_processes,
                group_id=self.owner.group_id,
                environment=env,
                conda_exe=conda_exe,
                server_ip=server_ip,
            )
        else:
            interpreter = self.tool_specification.execution_settings["executable"]
            python_exe = interpreter if interpreter else resolve_python_interpreter(self._settings)
            commands = self.make_python_basic_console_commands(cmdline_args)
            alias = f"python {' '.join([self.tool_specification.main_prgm] + cmdline_args)}"
            self.exec_mngr = PythonPersistentExecutionManager(
                self._logger, [python_exe], commands, alias, self._kill_completed_processes, self.owner.group_id
            )

    @staticmethod
    def _make_exec_code(fp, full_fp):
        """Returns a list of lines of code to run this tool instance in a Python interactive session.

        Args:
            fp (str): relative path to main Python file
            full_fp (str): absolute path to the main Python file

        Returns:
            list of str: lines of code
        """
        globals_dict = "globals_dict = globals()"
        update_globals_dict = f'globals_dict.update({{"__file__": "{full_fp}", "__name__": "__main__"}})'
        compile_and_exec = (
            f"with open('{fp}', 'rb') as f: exec(compile(f.read(), '{fp}', 'exec'), globals_dict, globals_dict)"
            + os.linesep
        )
        return [globals_dict, update_globals_dict, compile_and_exec]

    def execute(self):
        """Executes a prepared instance."""
        ret = self.exec_mngr.run_until_complete()
        if ret != 0:
            try:
                return_msg = self.tool_specification.return_codes[ret]
                self._logger.msg_error.emit(f"\t<b>{return_msg}</b> [exit code:{ret}]")
            except KeyError:
                self._logger.msg_error.emit(f"\tUnknown return code ({ret})")
        self._update_killed()
        self.exec_mngr = None
        return ret


class ExecutableToolInstance(ToolInstance):
    """Class for Executable Tool instances."""

    def prepare(self, args):
        """See base class."""
        if not self.tool_specification.main_prgm:  # Run command
            cmd = self.tool_specification.execution_settings["cmd"].split()  # Convert str to list
            shell = self.tool_specification.execution_settings["shell"]
            if not shell:
                # If shell is not given (empty str), The first item in cmd list will be considered as self.program.
                # The rest of the cmd list will be considered as cmd line args
                self.program = cmd.pop(0)
            else:
                # If shell is given, the shell will be set as self.program and the cmd list will be considered
                # as cmd line args
                if shell == "bash":
                    self.program = "sh"
                else:
                    self.program = shell
            if self.program in {"cmd.exe", "cmd"}:
                # If cmd.exe shell is not given the /C flag, it will just open cmd.exe in the Execution Log
                if "/C" not in cmd:
                    cmd = ["/C"] + cmd
            # The final command line args list (self.args) consists of three (3) parts:
            # 1. The first part is the cmd list
            # 2. The second part is the Tool Specification cmd line args
            # 3. The third part is the Tool cmd line args
            self.args = cmd + self.tool_specification.cmdline_args + args
        else:  # Run main program file
            main_program_file = os.path.join(self.basedir, self.tool_specification.main_prgm)
            if os.path.isfile(main_program_file):
                if sys.platform != "win32":
                    self.program = "sh"
                    self.args.append(main_program_file)
                else:
                    self.program = main_program_file
                self.append_cmdline_args(args)
            else:
                raise RuntimeError(f"main program file {main_program_file} does not exist.")
        self.exec_mngr = ProcessExecutionManager(self._logger, self.program, self.args, workdir=self.basedir)

    def execute(self):
        """Executes a prepared instance."""
        ret = self.exec_mngr.run_until_complete()
        if ret != 0:
            try:
                return_msg = self.tool_specification.return_codes[ret]
                self._logger.msg_error.emit(f"\t<b>{return_msg}</b> [exit code:{ret}]")
            except KeyError:
                self._logger.msg_error.emit(f"\tUnknown return code ({ret})")
        self._update_killed()
        self.exec_mngr = None
        return ret



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""This module contains utilities for Data Connection."""
import sys
import urllib.parse
from spine_engine.utils.serialization import deserialize_path
from spine_items.utils import convert_url_to_safe_string


def restore_database_references(references_list, credentials_dict, project_dir):
    """Restores data from serialized database references.

    Args:
        references_list (list of dict): serialized database references
        credentials_dict (dict): mapping from safe URL to (username, password) tuple
        project_dir (str): path to project directory

    Returns:
        list of dict: deserialized database references
    """
    db_references = []
    for reference_dict in references_list:
        if isinstance(reference_dict, str):
            # legacy db reference
            url = urllib.parse.urlparse(reference_dict)
            dialect = _dialect_from_scheme(url.scheme)
            path = url.path
            if dialect == "sqlite" and sys.platform == "win32":
                # Remove extra '/' from file path on Windows.
                path = path[1:]
            db_reference = {
                "dialect": dialect,
                "host": url.hostname,
                "port": url.port,
                "database": path,
            }
        else:
            db_reference = dict(reference_dict)
        if db_reference["dialect"] == "sqlite":
            db_reference["database"] = deserialize_path(db_reference["database"], project_dir)
        db_reference["username"], db_reference["password"] = credentials_dict.get(
            convert_url_to_safe_string(db_reference), (None, None)
        )
        db_references.append(db_reference)
    return db_references


def _dialect_from_scheme(scheme):
    """Parses dialect from URL scheme.

    Args:
        scheme (str): URL scheme

    Returns:
        str: dialect name
    """
    return scheme.split("+")[0]



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Contains unit tests for the ImportEditorWindow class. """
import unittest
from unittest import mock
from PySide6.QtCore import QSettings
from PySide6.QtWidgets import QApplication, QWidget
from spine_items.importer.widgets.import_editor_window import ImportEditorWindow
from spinedb_api.spine_io.importers.csv_reader import CSVReader


class TestImportEditorWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_closeEvent(self):
        spec = mock.NonCallableMagicMock()
        spec.name = "spec_name"
        spec.description = "spec_desc"
        spec.mapping = {"source_type": CSVReader.__name__}
        toolbox = QWidget()
        toolbox.qsettings = mock.MagicMock(return_value=QSettings(toolbox))
        toolbox.restore_and_activate = mock.MagicMock()
        widget = ImportEditorWindow(toolbox, spec)
        QApplication.processEvents()  # Let QTimer call ImportEditorWindow.start_ui().
        widget._app_settings = mock.NonCallableMagicMock()
        widget.close()
        widget._app_settings.beginGroup.assert_called_once_with("mappingPreviewWindow")
        widget._app_settings.endGroup.assert_called_once_with()
        qsettings_save_calls = widget._app_settings.setValue.call_args_list
        self.assertEqual(len(qsettings_save_calls), 8)
        saved_dict = {saved[0][0]: saved[0][1] for saved in qsettings_save_calls}
        self.assertEqual(len(saved_dict), 8)
        for key in (
            "windowSize",
            "windowPosition",
            "windowState",
            "windowMaximized",
            "n_screens",
            "splitter_source_listState",
            "splitter_source_data_mappingsState",
            "splitterState",
        ):
            with self.subTest(key=key):
                self.assertIn(key, saved_dict)
        toolbox.deleteLater()


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Helper functions for Importer's unit tests."""


def append_mapping(test_case, mappings_model, name, table_index):
    row_count = mappings_model.rowCount(table_index)
    test_case.assertTrue(mappings_model.insertRow(row_count, table_index))
    test_case.assertEqual(mappings_model.rowCount(table_index), row_count + 1)
    mappings_model.rename_mapping(table_index.row(), row_count, name)
    test_case.assertEqual(mappings_model.index(row_count, 0, table_index).data(), name)


def append_source_table_with_mappings(
    test_case, mappings_model, table_name, first_mapping_name, additional_mapping_names
):
    original_row_count = mappings_model.rowCount()
    root_mapping = mappings_model.create_default_mapping()
    mappings_model.append_new_table_with_mapping(table_name, root_mapping)
    test_case.assertEqual(mappings_model.rowCount(), original_row_count + 1)
    table_row = original_row_count - (1 if mappings_model.has_empty_source_table_row() else 0)
    table_index = mappings_model.index(table_row, 0)
    test_case.assertEqual(table_index.data(), table_name)
    test_case.assertEqual(mappings_model.rowCount(table_index), 1)
    mappings_model.rename_mapping(table_row, 0, first_mapping_name)
    test_case.assertEqual(mappings_model.index(0, 0, table_index).data(), first_mapping_name)
    for name in additional_mapping_names:
        append_mapping(test_case, mappings_model, name, table_index)
    return table_index



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ToolExecutable item."""
from multiprocessing import Lock
import pathlib
from queue import Queue
import sys
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from PySide6.QtCore import QCoreApplication
from spine_engine.execution_managers.persistent_execution_manager import kill_persistent_processes
from spine_engine.project_item.project_item_resource import CmdLineArg, file_resource
from spine_engine.utils.queue_logger import QueueLogger
from spine_items.tool.executable_item import ExecutableItem, _count_files_and_dirs
from spine_items.tool.tool_specifications import PythonTool, ToolSpecification


class TestToolExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "Tool")

    def test_from_dict(self):
        """Tests that from_dict creates an ExecutableItem."""
        mock_settings = _MockSettings()
        item_dict = {
            "type": "Tool",
            "description": "",
            "x": 0,
            "y": 0,
            "specification": "Python Tool",
            "execute_in_work": True,
            "cmd_line_args": [{"type": "literal", "arg": "a"}, {"type": "literal", "arg": "b"}],
            "options": {},
        }
        script_dir = pathlib.Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = self._write_output_script(script_dir)
        script_files = [script_file_name]
        python_tool_spec = PythonTool(
            name="Python Tool",
            tooltype="Python",
            path=str(script_dir),
            includes=script_files,
            settings=mock_settings,
            logger=mock.MagicMock(),
        )
        specs_in_project = {"Tool": {"Python Tool": python_tool_spec}}
        temp_project_dir = str(pathlib.Path(self._temp_dir.name, "project"))
        item = ExecutableItem.from_dict(
            item_dict,
            name="T",
            project_dir=temp_project_dir,
            app_settings=mock_settings,
            specifications=specs_in_project,
            logger=mock.MagicMock(),
        )
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("Tool", item.item_type())
        self.assertTrue(item._tool_specification.name, "Python Tool")
        self.assertEqual("some_work_dir", item._work_dir)
        self.assertEqual([CmdLineArg("a"), CmdLineArg("b")], item._cmd_line_args)
        # Test that the item is not created if "appSettings/workDir" key is missing from qsettings
        item = ExecutableItem.from_dict(
            item_dict,
            name="T",
            project_dir=temp_project_dir,
            app_settings=_EmptyMockSettings(),
            specifications=specs_in_project,
            logger=mock.MagicMock(),
        )
        self.assertIsInstance(item, ExecutableItem)
        self.assertIsNone(item._work_dir, "")
        # This time the project dict does not have any specifications
        item = ExecutableItem.from_dict(
            item_dict,
            name="T",
            project_dir=temp_project_dir,
            app_settings=mock_settings,
            specifications={},
            logger=mock.MagicMock(),
        )
        self.assertIsInstance(item, ExecutableItem)
        self.assertIsNone(item._tool_specification)
        # Modify item_dict
        item_dict["execute_in_work"] = False
        item = ExecutableItem.from_dict(
            item_dict,
            name="T",
            project_dir=temp_project_dir,
            app_settings=mock_settings,
            specifications=specs_in_project,
            logger=mock.MagicMock(),
        )
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("Tool", item.item_type())
        self.assertEqual([CmdLineArg("a"), CmdLineArg("b")], item._cmd_line_args)
        self.assertIsNone(item._work_dir)
        # Modify item_dict
        item_dict["specification"] = ""
        item = ExecutableItem.from_dict(
            item_dict,
            name="T",
            project_dir=temp_project_dir,
            app_settings=mock_settings,
            specifications=specs_in_project,
            logger=mock.MagicMock(),
        )
        self.assertIsInstance(item, ExecutableItem)

    def test_ready_to_execute(self):
        logger = mock.MagicMock()
        executable = ExecutableItem(
            "executable name",
            work_dir=str(pathlib.Path(self._temp_dir.name, "work")),
            tool_specification=None,
            cmd_line_args=[],
            options={},
            kill_completed_processes=False,
            log_process_output=False,
            group_id=None,
            project_dir=self._temp_dir.name,
            logger=logger,
        )
        self.assertFalse(executable.ready_to_execute(settings={}))

    def test_execute_archives_output_files(self):
        script_dir = pathlib.Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = self._write_output_script(script_dir)
        script_files = [script_file_name]
        output_files = ["out.dat", "subdir/out.txt"]
        app_settings = _MockSettings()
        logger = mock.MagicMock()
        tool_specification = PythonTool(
            "Python tool", "Python", str(script_dir), script_files, app_settings, logger, outputfiles=output_files
        )
        work_dir = pathlib.Path(self._temp_dir.name, "work")
        work_dir.mkdir()
        executable = ExecutableItem(
            "Create files", str(work_dir), tool_specification, [], {}, False, False, None, self._temp_dir.name, logger
        )
        executable.execute([], [], Lock())
        while executable._tool_instance is not None:
            QCoreApplication.processEvents()
        archives = list(pathlib.Path(executable._data_dir, "output").iterdir())
        self.assertEqual(len(archives), 1)
        self.assertNotEqual(archives[0].name, "failed")
        self.assertTrue(pathlib.Path(archives[0], "out.dat").exists())
        self.assertTrue(pathlib.Path(archives[0], "subdir", "out.txt").exists())
        kill_persistent_processes()

    def test_execute_logs_messages(self):
        script_dir = pathlib.Path(self._temp_dir.name, "scripts")
        script_dir.mkdir()
        script_file_name = "script.py"
        file_path = pathlib.Path(script_dir, "script.py")
        with open(file_path, "w") as script_file:
            script_file.writelines(["print('hello')\n", "raise ValueError('foo')\n"])
        script_files = [script_file_name]
        app_settings = _MockSettings()
        item_name = "Logs stuff"
        logger = QueueLogger(Queue(), item_name, None, [])
        tool_specification = PythonTool("Python tool", "Python", str(script_dir), script_files, app_settings, logger)
        work_dir = pathlib.Path(self._temp_dir.name, "work")
        work_dir.mkdir()
        executable = ExecutableItem(
            item_name, str(work_dir), tool_specification, [], {}, False, True, None, self._temp_dir.name, logger
        )
        executable.execute([], [], Lock())
        while executable._tool_instance is not None:
            QCoreApplication.processEvents()
        logs = list(pathlib.Path(executable._logs_dir).iterdir())
        self.assertEqual(len(logs), 1)
        with open(logs[0], "r") as f:
            lines = f.readlines()
        expected_stdout = [
            "### Spine execution log file\n",
            "### Item name: Logs stuff\n",
            "### Filter id: \n",
            "### Part: 1\n",
            "\n",
            "# Running python script.py\n",
            "hello\n",
        ]
        if sys.version_info.minor < 13:
            expected_stderr = [
                "Traceback (most recent call last):\n",
                '  File "<stdin>", line 3, in <module>\n',
                '  File "script.py", line 2, in <module>\n',
                "    raise ValueError('foo')\n",
                "ValueError: foo\n",
            ]
        else:
            expected_stderr = [
                "Traceback (most recent call last):\n",
                '  File "<stdin>", line 3, in <module>\n',
                "    with open('script.py', 'rb') as f: exec(compile(f.read(), 'script.py', 'exec'), globals_dict, globals_dict)\n",
                "                                       ~~~~^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n",
                '  File "script.py", line 2, in <module>\n',
                "    raise ValueError('foo')\n",
                "ValueError: foo\n",
            ]
        self.assertCountEqual(lines, expected_stdout + expected_stderr)
        kill_persistent_processes()

    def test_find_optional_input_files_without_wildcards(self):
        optional_file = pathlib.Path(self._temp_dir.name, "1.txt")
        optional_file.touch()
        pathlib.Path(self._temp_dir.name, "should_not_be_found.txt").touch()
        logger = mock.MagicMock()
        optional_input_files = ["1.txt", "does_not_exist.dat"]
        tool_specification = ToolSpecification(
            "spec name", "Python", self._temp_dir.name, [], None, logger, inputfiles_opt=optional_input_files
        )
        executable = ExecutableItem(
            "executable name",
            work_dir=self._temp_dir.name,
            tool_specification=tool_specification,
            cmd_line_args=[],
            options={},
            kill_completed_processes=False,
            log_process_output=False,
            group_id=None,
            project_dir=self._temp_dir.name,
            logger=logger,
        )
        resources = [file_resource("provider", str(optional_file))]
        file_paths = executable._find_optional_input_files(resources)
        expected = {"1.txt": [optional_file]}
        self.assertEqual(len(file_paths), len(expected))
        for label, expected_files in expected.items():
            paths = file_paths[label]
            self.assertEqual(len(paths), len(expected_files))
            for file_path, expected_path in zip(paths, expected_files):
                self.assertTrue(expected_path.samefile(pathlib.Path(file_path)))

    def test_find_optional_input_files_with_wildcards(self):
        optional_file1 = pathlib.Path(self._temp_dir.name, "1.txt")
        optional_file1.touch()
        optional_file2 = pathlib.Path(self._temp_dir.name, "2.txt")
        optional_file2.touch()
        pathlib.Path(self._temp_dir.name, "should_not_be_found.jpg").touch()
        logger = mock.MagicMock()
        optional_input_files = ["*.txt"]
        tool_specification = ToolSpecification(
            "spec name", "Python", self._temp_dir.name, [], None, logger, inputfiles_opt=optional_input_files
        )
        executable = ExecutableItem(
            "executable name",
            work_dir=self._temp_dir.name,
            tool_specification=tool_specification,
            cmd_line_args=[],
            options={},
            kill_completed_processes=False,
            log_process_output=False,
            group_id=None,
            project_dir=self._temp_dir.name,
            logger=logger,
        )
        resources = [file_resource("provider", str(optional_file1)), file_resource("provider", str(optional_file2))]
        file_paths = executable._find_optional_input_files(resources)
        expected = {"*.txt": [optional_file1, optional_file2]}
        self.assertEqual(len(file_paths), len(expected))
        for label, expected_files in expected.items():
            paths = file_paths[label]
            self.assertEqual(len(paths), len(expected_files))
            for file_path, expected_path in zip(paths, expected_files):
                self.assertTrue(expected_path.samefile(pathlib.Path(file_path)))

    def test_find_optional_input_files_in_sub_directory(self):
        pathlib.Path(self._temp_dir.name, "subdir").mkdir()
        optional_file1 = pathlib.Path(self._temp_dir.name, "subdir", "1.txt")
        optional_file1.touch()
        optional_file2 = pathlib.Path(self._temp_dir.name, "subdir", "data.dat")
        optional_file2.touch()
        pathlib.Path(self._temp_dir.name, "should_not_be_found.jpg").touch()
        logger = mock.MagicMock()
        optional_input_files = ["subdir/*.txt", "subdir/data.dat"]
        tool_specification = ToolSpecification(
            "spec name", "Python", self._temp_dir.name, [], None, logger, inputfiles_opt=optional_input_files
        )
        executable = ExecutableItem(
            "executable name",
            work_dir=self._temp_dir.name,
            tool_specification=tool_specification,
            cmd_line_args=[],
            options={},
            kill_completed_processes=False,
            log_process_output=False,
            group_id=None,
            project_dir=self._temp_dir.name,
            logger=logger,
        )
        resources = [file_resource("provider", str(optional_file1)), file_resource("provider", str(optional_file2))]
        file_paths = executable._find_optional_input_files(resources)
        expected = {"subdir/*.txt": [optional_file1], "subdir/data.dat": [optional_file2]}
        self.assertEqual(len(file_paths), len(expected))
        for label, expected_files in expected.items():
            paths = file_paths[label]
            self.assertEqual(len(paths), len(expected_files))
            for file_path, expected_path in zip(paths, expected_files):
                self.assertTrue(expected_path.samefile(pathlib.Path(file_path)))

    def test_output_resources_forward(self):
        logger = mock.MagicMock()
        tool_specification = PythonTool(
            name="Python Tool",
            tooltype="Python",
            path=self._temp_dir.name,
            includes=["script.py"],
            settings=None,
            logger=mock.MagicMock(),
            outputfiles=["results.gdx", "report.txt"],
        )
        executable = ExecutableItem(
            "name", self._temp_dir.name, tool_specification, [], {}, False, False, None, self._temp_dir.name, logger
        )
        output_dir = pathlib.Path(executable._data_dir, "output")
        output_dir.mkdir()
        out_file_1 = output_dir / "results.gdx"
        out_file_1.touch()
        out_file_2 = output_dir / "report.txt"
        out_file_2.touch()
        with mock.patch("spine_items.tool.output_resources.find_last_output_files") as mock_find_last_output_files:
            mock_find_last_output_files.return_value = {
                "results.gdx": [str(out_file_1)],
                "report.txt": [str(out_file_2)],
            }
            resources = executable._output_resources_forward()
            mock_find_last_output_files.assert_called_once()
            self.assertIsInstance(resources, list)
            self.assertEqual(2, len(resources))
            resource_paths = [r.path for r in resources]
            self.assertTrue(any(out_file_1.samefile(p) for p in resource_paths))
            self.assertTrue(any(out_file_2.samefile(p) for p in resource_paths))

    def test_stop_execution(self):
        logger = mock.MagicMock()
        tool_specification = PythonTool(
            name="Python Tool",
            tooltype="Python",
            path=self._temp_dir.name,
            includes=["script.py"],
            settings=None,
            logger=mock.MagicMock(),
            outputfiles=["results.gdx", "report.txt"],
        )
        executable = ExecutableItem(
            "name", self._temp_dir.name, tool_specification, [], {}, False, False, None, self._temp_dir.name, logger
        )
        executable._tool_instance = executable._tool_specification.create_tool_instance(
            self._temp_dir.name, False, logger, mock.MagicMock()
        )
        executable._tool_instance.exec_mngr = mock.MagicMock()
        executable.stop_execution()
        self.assertIsNone(executable._tool_instance)

    def test_count_files_and_dirs(self):
        """Tests protected function in tool/executable_item.py."""
        paths = ["data/a.txt", "data/output", "data/input_dir/", "inc/b.txt", "directory/"]  # 3 files, 2 dirs
        n_dir, n_files = _count_files_and_dirs(paths)
        self.assertEqual(2, n_dir)
        self.assertEqual(3, n_files)

    @staticmethod
    def _write_output_script(script_dir):
        file_path = pathlib.Path(script_dir, "script.py")
        with open(file_path, "w") as script_file:
            script_file.writelines(
                [
                    "from pathlib import Path\n",
                    "Path('out.dat').touch()\n",
                    "Path('subdir').mkdir(exist_ok=True)\n",
                    "Path('subdir', 'out.txt').touch()\n",
                ]
            )
        return "script.py"


class _EmptyMockSettings:
    @staticmethod
    def value(key, defaultValue=None):
        return {None: None}.get(key, defaultValue)


class _MockSettings:
    @staticmethod
    def value(key, defaultValue=None):
        return {
            "appSettings/pythonPath": sys.executable,
            "appSettings/usePythonKernel": "0",  # Don't use Jupyter Console
            "appSettings/workDir": "some_work_dir",
        }.get(key, defaultValue)


if __name__ == "__main__":
    unittest.main()
